"""API route implementations for Docler."""

from __future__ import annotations

import mimetypes
import tempfile
from typing import TYPE_CHECKING, Annotated, Any

import anyenv
from fastapi import Body, File, Form, HTTPException, Query
from pydantic import TypeAdapter
import upath

from docler.configs.converter_configs import ConverterConfig
from docler.models import PageMetadata  # noqa: TC001
from docler.pdf_utils import decrypt_pdf, get_pdf_info


if TYPE_CHECKING:
    from fastapi import UploadFile

    from docler.configs.chunker_configs import ChunkerConfig


config_adapter = TypeAdapter[ConverterConfig](ConverterConfig)


async def convert_document(
    file: Annotated[UploadFile, File(description="The document file to convert")],
    config: Annotated[str, Form(description="Converter configuration JSON")],
    pdf_password: Annotated[
        str | None, Form(description="Password for encrypted PDF files")
    ] = None,
):
    """Convert a document file to markdown using specified converter configuration."""
    # Parse the JSON config string manually

    try:
        config_dict = anyenv.load_json(config)
        parsed = config_adapter.validate_python(config_dict)
    except (anyenv.JsonLoadError, ValueError) as e:
        raise HTTPException(status_code=400, detail=f"Invalid config JSON: {e}")  # noqa: B904

    content = await file.read()

    # Handle PDF decryption if needed
    if file.content_type == "application/pdf" or (
        file.filename and file.filename.lower().endswith(".pdf")
    ):
        # Check if PDF is encrypted
        try:
            pdf_info = get_pdf_info(content)
            if pdf_info.is_encrypted:
                if pdf_password is None:
                    # Try empty password first
                    try:
                        content = decrypt_pdf(content, None)
                    except ValueError as e:
                        if "requires a password" in str(e):
                            raise HTTPException(
                                status_code=400,
                                detail=(
                                    "PDF is encrypted but no password provided. "
                                    "Please provide pdf_password parameter."
                                ),
                            ) from e
                        raise HTTPException(
                            status_code=400, detail=f"Failed to decrypt PDF: {e}"
                        ) from e
                else:
                    # Decrypt the PDF with provided password
                    try:
                        content = decrypt_pdf(content, pdf_password)
                    except ValueError as e:
                        if "Incorrect password" in str(e):
                            raise HTTPException(  # noqa: B904
                                status_code=401, detail="Incorrect PDF password"
                            )
                        raise HTTPException(
                            status_code=400, detail=f"Failed to decrypt PDF: {e}"
                        ) from e
        except ValueError as e:
            if "encrypted" not in str(e).lower():
                # If it's not an encryption issue, let it pass through to the converter
                pass

    with tempfile.NamedTemporaryFile(
        suffix=f"_{file.filename}", delete=False
    ) as temp_file:
        temp_path = temp_file.name
        temp_file.write(content)

    try:
        mime_type, _ = mimetypes.guess_type(file.filename or "")
        if not mime_type:
            mime_type = file.content_type

        # Create converter from config
        converter = parsed.get_provider()
        document = await converter.convert_file(temp_path)

        # Always include images as base64
        for image in document.images:
            if isinstance(image.content, bytes):
                # Convert bytes to base64 string
                image.content = image.to_base64()
    except Exception as e:
        if not isinstance(e, HTTPException):
            raise HTTPException(
                status_code=500, detail=f"Error during document conversion: {e!s}"
            ) from None
        raise
    else:
        return document
    finally:
        # Clean up the temporary file
        path = upath.UPath(temp_path)
        path.unlink(missing_ok=True)


async def chunk_document(
    file: Annotated[UploadFile, File(description="The document file to chunk")],
    converter_config: Annotated[
        ConverterConfig,
        Body(default={"type": "marker"}, description="Converter configuration"),
    ],
    chunker_config: Annotated[
        ChunkerConfig,
        Body(default={"type": "markdown"}, description="Chunker configuration"),
    ],
    pdf_password: str | None = Query(
        default=None,
        description="Password for encrypted PDF files",
    ),
):
    """Convert and chunk a document file using specified configurations.

    Args:
        file: The document file to convert and chunk
        converter_config: Configuration for the document converter
        chunker_config: Configuration for the text chunker
        pdf_password: Password for encrypted PDF files

    Returns:
        JSON response with the chunked document
    """
    content = await file.read()

    # Handle PDF decryption if needed
    if file.content_type == "application/pdf" or (
        file.filename and file.filename.lower().endswith(".pdf")
    ):
        # Check if PDF is encrypted
        try:
            pdf_info = get_pdf_info(content)
            if pdf_info.is_encrypted:
                if pdf_password is None:
                    # Try empty password first
                    try:
                        content = decrypt_pdf(content, None)
                    except ValueError as e:
                        if "requires a password" in str(e):
                            raise HTTPException(
                                status_code=400,
                                detail="PDF is encrypted but no password provided.",
                            ) from e
                        raise HTTPException(
                            status_code=400, detail=f"Failed to decrypt PDF: {e}"
                        ) from e
                else:
                    # Decrypt the PDF with provided password
                    try:
                        content = decrypt_pdf(content, pdf_password)
                    except ValueError as e:
                        if "Incorrect password" in str(e):
                            raise HTTPException(  # noqa: B904
                                status_code=401, detail="Incorrect PDF password"
                            )
                        raise HTTPException(
                            status_code=400, detail=f"Failed to decrypt PDF: {e}"
                        ) from e
        except ValueError as e:
            if "encrypted" not in str(e).lower():
                # If it's not an encryption issue, let it pass through to the converter
                pass

    with tempfile.NamedTemporaryFile(
        suffix=f"_{file.filename}", delete=False
    ) as temp_file:
        temp_path = temp_file.name
        temp_file.write(content)

    try:
        mime_type, _ = mimetypes.guess_type(file.filename or "")
        if not mime_type:
            mime_type = file.content_type

        # Get converter and chunker instances
        converter = converter_config.get_provider()
        chunker = chunker_config.get_provider()

        # Convert the document
        document = await converter.convert_file(temp_path)

        # Chunk the document
        chunked_document = await chunker.chunk(document)

        # Always include images as base64
        for image in chunked_document.images:
            if isinstance(image.content, bytes):
                # Convert bytes to base64 string
                image.content = image.to_base64()
        for chunk in chunked_document.chunks:
            for image in chunk.images:
                if isinstance(image.content, bytes):
                    # Convert bytes to base64 string
                    image.content = image.to_base64()
    except Exception as e:
        if not isinstance(e, HTTPException):
            raise HTTPException(
                status_code=500,
                detail=f"Error during document conversion or chunking: {e!s}",
            ) from e
        raise
    else:
        return chunked_document
    finally:
        # Clean up the temporary file
        path = upath.UPath(temp_path)
        path.unlink(missing_ok=True)


async def list_converters():
    """List all available converters."""
    from docler.converters.registry import ConverterRegistry

    registry = ConverterRegistry.create_default()

    converters = [
        {
            "name": converter.NAME,
            "supported_mime_types": list(converter.SUPPORTED_MIME_TYPES),
            "config_type": converter.__class__.Config.__name__,
            "config_schema": converter.__class__.Config.model_json_schema(),
        }
        for converter in registry._converters
    ]

    return {"converters": converters}


async def list_chunkers():
    """List all available chunkers."""
    from docler.chunkers.base import TextChunker

    chunker_classes = TextChunker[Any].get_available_providers()
    chunkers = [
        {
            "name": chunker_class.NAME,
            "config_type": chunker_class.Config.__name__,
            "config_schema": chunker_class.Config.model_json_schema(),
        }
        for chunker_class in chunker_classes
        if chunker_class.has_required_packages()
    ]

    return {"chunkers": chunkers}


async def get_pdf_metadata(
    file: Annotated[UploadFile, File(description="The PDF file to analyze")],
    pdf_password: Annotated[
        str | None, Form(description="Password for encrypted PDF files")
    ] = None,
) -> PageMetadata:
    """Get PDF metadata including page count and document information.

    Args:
        file: The PDF file to analyze
        pdf_password: Password for encrypted PDF files

    Returns:
        PageMetadata containing document information

    Raises:
        HTTPException: If file is invalid or processing fails
    """
    # Validate file type
    if file.content_type != "application/pdf":
        raise HTTPException(status_code=400, detail="Only PDF files are supported")

    # Validate file size (100MB limit for metadata extraction)
    if file.size and file.size > 100 * 1024 * 1024:
        raise HTTPException(
            status_code=400, detail="File size too large. Maximum size is 100MB."
        )

    try:
        content = await file.read()
        metadata = get_pdf_info(content, pdf_password)
    except ValueError as e:
        if "Incorrect password" in str(e):
            raise HTTPException(status_code=401, detail="Incorrect PDF password") from e
        raise HTTPException(
            status_code=400, detail=f"Failed to process PDF: {e!s}"
        ) from e
    except Exception as e:
        raise HTTPException(
            status_code=500, detail=f"Internal error processing PDF: {e!s}"
        ) from e
    else:
        return metadata



"""Configuration models for Docler components."""

from __future__ import annotations

from docler.configs.annotator_configs import (
    AIDocumentAnnotatorConfig,
    AIImageAnnotatorConfig,
    AnnotatorConfig,
    BaseAnnotatorConfig,
    DEFAULT_DOC_PROMPT_TEMPLATE,
    DEFAULT_DOC_SYSTEM_PROMPT,
    DEFAULT_IMAGE_PROMPT_TEMPLATE,
    DEFAULT_IMAGE_SYSTEM_PROMPT,
)
from docler.configs.chunker_configs import (
    AiChunkerConfig,
    BaseChunkerConfig,
    ChunkerConfig,
    LlamaIndexChunkerConfig,
    MarkdownChunkerConfig,
    ChunkerShorthand,
    DEFAULT_CHUNKER_USER_TEMPLATE,
    DEFAULT_CHUNKER_SYSTEM_PROMPT,
)
from docler.configs.converter_configs import (
    ConverterShorthand,
    AzureConfig,
    AzureFeatureFlag,
    AzureModel,
    BaseConverterConfig,
    ConverterConfig,
    DataLabConfig,
    DoclingConverterConfig,
    DoclingEngine,
    LlamaParseConfig,
    LlamaParseMode,
    LLMConverterConfig,
    MarkerConfig,
    MarkItDownConfig,
    MistralConfig,
    UpstageConfig,
)
from docler.configs.embedding_configs import (
    EmbeddingShorthand,
    BaseEmbeddingConfig,
    BGEEmbeddingConfig,
    EmbeddingConfig,
    LiteLLMEmbeddingConfig,
    LiteLLMInputType,
    OpenAIEmbeddingConfig,
    OpenAIEmbeddingModel,
    SentenceTransformerEmbeddingConfig,
    SentenceTransformerModel,
)
from docler.configs.file_db_configs import (
    ComponentBasedConfig,
    DatabaseShorthand,
    OpenAIChunkingStrategy,
    FileDatabaseConfig,
    FileDatabaseConfigUnion,
    OpenAIFileDatabaseConfig,
    resolve_database_config,
)
from docler.configs.processor_configs import (
    BaseProcessorConfig,
    DEFAULT_PROOF_READER_PROMPT_TEMPLATE,
    DEFAULT_PROOF_READER_SYSTEM_PROMPT,
    LLMProofReaderConfig,
    ProcessorConfig,
)
from docler.configs.vector_db_configs import (
    VectorDBShorthand,
    BaseVectorStoreConfig,
    ChromaConfig,
    KdbAiConfig,
    Metric,
    PineconeCloud,
    PineconeConfig,
    PineconeRegion,
    QdrantConfig,
    VectorStoreConfig,
)

__all__ = [
    "DEFAULT_CHUNKER_SYSTEM_PROMPT",
    "DEFAULT_CHUNKER_USER_TEMPLATE",
    "DEFAULT_DOC_PROMPT_TEMPLATE",
    "DEFAULT_DOC_SYSTEM_PROMPT",
    "DEFAULT_IMAGE_PROMPT_TEMPLATE",
    "DEFAULT_IMAGE_SYSTEM_PROMPT",
    "DEFAULT_PROOF_READER_PROMPT_TEMPLATE",
    "DEFAULT_PROOF_READER_SYSTEM_PROMPT",
    "AIDocumentAnnotatorConfig",
    "AIImageAnnotatorConfig",
    "AiChunkerConfig",
    "AnnotatorConfig",
    "AzureConfig",
    "AzureFeatureFlag",
    "AzureModel",
    "BGEEmbeddingConfig",
    "BaseAnnotatorConfig",
    "BaseChunkerConfig",
    "BaseConverterConfig",
    "BaseEmbeddingConfig",
    "BaseProcessorConfig",
    "BaseVectorStoreConfig",
    "ChromaConfig",
    "ChunkerConfig",
    "ChunkerShorthand",
    "ComponentBasedConfig",
    "ConverterConfig",
    "ConverterShorthand",
    "DataLabConfig",
    "DatabaseShorthand",
    "DoclingConverterConfig",
    "DoclingEngine",
    "EmbeddingConfig",
    "EmbeddingShorthand",
    "FileDatabaseConfig",
    "FileDatabaseConfigUnion",
    "KdbAiConfig",
    "LLMConverterConfig",
    "LLMProofReaderConfig",
    "LiteLLMEmbeddingConfig",
    "LiteLLMInputType",
    "LlamaIndexChunkerConfig",
    "LlamaParseConfig",
    "LlamaParseMode",
    "MarkItDownConfig",
    "MarkdownChunkerConfig",
    "MarkerConfig",
    "Metric",
    "MistralConfig",
    "OpenAIChunkingStrategy",
    "OpenAIEmbeddingConfig",
    "OpenAIEmbeddingModel",
    "OpenAIFileDatabaseConfig",
    "PineconeCloud",
    "PineconeConfig",
    "PineconeRegion",
    "ProcessorConfig",
    "QdrantConfig",
    "SentenceTransformerEmbeddingConfig",
    "SentenceTransformerModel",
    "UpstageConfig",
    "VectorDBShorthand",
    "VectorStoreConfig",
    "resolve_database_config",
]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from mkdown import Document
from schemez import Schema

from docler.common_types import DEFAULT_PROOF_READER_MODEL
from docler.configs.chunker_configs import (
    AiChunkerConfig,
    LlamaIndexChunkerConfig,
    MarkdownChunkerConfig,
    TokenAwareChunkerConfig,
)
from docler.configs.processor_configs import (
    DEFAULT_PROOF_READER_PROMPT_TEMPLATE,
    DEFAULT_PROOF_READER_SYSTEM_PROMPT,
    LLMProofReaderConfig,
)
from docler.diffs import generate_all_diffs
from docler.processors.base import DocumentProcessor
from docler.utils import add_line_numbers


if TYPE_CHECKING:
    from docler.configs.chunker_configs import ChunkerConfig
    from docler.configs.file_db_configs import ChunkerShorthand


class LineCorrection(Schema):
    """A correction to apply to a specific line."""

    line_number: int
    """The line number to correct (1-based)."""

    corrected: str
    """The corrected text."""


def apply_corrections(
    text: str, corrections: list[LineCorrection]
) -> tuple[str, set[int]]:
    """Apply corrections to the original text.

    Args:
        text: Original text to apply corrections to
        corrections: List of line corrections

    Returns:
        Tuple containing (corrected text, set of corrected line indices)
    """
    lines = text.splitlines()
    corrections.sort(key=lambda c: c.line_number, reverse=True)
    corrected_lines = set()
    for correction in corrections:
        line_idx = correction.line_number - 1
        if 0 <= line_idx < len(lines) and line_idx not in corrected_lines:
            lines[line_idx] = correction.corrected
            corrected_lines.add(line_idx)

    return "\n".join(lines), corrected_lines


def resolve_chunker_config(
    config: ChunkerConfig | ChunkerShorthand | None, model: str
) -> ChunkerConfig | None:
    """Resolve chunker configuration from shorthand or create default.

    Args:
        config: Shorthand string, configuration object, or None
        model: Model to use for default token-aware chunker

    Returns:
        Fully resolved configuration object
    """
    if isinstance(config, str):
        match config:
            case "markdown":
                return MarkdownChunkerConfig()
            case "llamaindex":
                return LlamaIndexChunkerConfig()
            case "ai":
                return AiChunkerConfig()
            case "token_aware":
                return TokenAwareChunkerConfig(model=model)

    # Return the config if it's already a config object
    return config


class LLMProofReader(DocumentProcessor[LLMProofReaderConfig]):
    """LLM-based proof-reader that improves OCR output using line-based corrections."""

    Config = LLMProofReaderConfig
    REQUIRED_PACKAGES: ClassVar = {"llmling-agent"}
    NAME = "proof_reading"

    def __init__(
        self,
        model: str | None = None,
        system_prompt: str | None = None,
        prompt_template: str | None = None,
        chunker: ChunkerConfig | ChunkerShorthand | None = None,
        include_diffs: bool = True,
        add_metadata_only: bool = False,
    ):
        """Initialize LLM document proof-reader.

        Args:
            model: LLM model to use
            system_prompt: Custom system prompt
            prompt_template: Custom prompt template
            chunker: Custom chunker configuration or shorthand
                     (if None, uses TokenAwareChunker with this model)
            include_diffs: Whether to include diffs in metadata
            add_metadata_only: If True, only add metadata without modifying content
        """
        self.model = model or DEFAULT_PROOF_READER_MODEL
        self.system_prompt = system_prompt or DEFAULT_PROOF_READER_SYSTEM_PROMPT
        self.prompt_template = prompt_template or DEFAULT_PROOF_READER_PROMPT_TEMPLATE
        self.include_diffs = include_diffs
        self.add_metadata_only = add_metadata_only

        # Resolve chunker config
        self.chunker_config = resolve_chunker_config(chunker, self.model)

    async def process(self, doc: Document) -> Document:
        """Process document using line-based corrections."""
        from llmling_agent import Agent

        agent = Agent[None](model=self.model, system_prompt=self.system_prompt)

        corrections = []
        if self.chunker_config:
            # Process with chunking
            chunker = self.chunker_config.get_provider()
            temp_doc = Document(content=doc.content, source_path=doc.source_path)
            chunks = await chunker.split(temp_doc)

            for chunk in chunks:
                numbered_text = chunk.to_numbered_text()
                user_prompt = self.prompt_template.format(chunk_text=numbered_text)

                chunk_corrections = await agent.talk.extract_multiple(
                    text=numbered_text,
                    as_type=LineCorrection,
                    prompt=user_prompt,
                    mode="structured",
                )
                corrections.extend(chunk_corrections)
        else:
            # Process the entire document at once
            numbered_text = add_line_numbers(doc.content)
            user_prompt = self.prompt_template.format(chunk_text=numbered_text)

            corrections = await agent.talk.extract_multiple(
                text=numbered_text,
                as_type=LineCorrection,
                prompt=user_prompt,
                mode="structured",
            )

        new_content, corrected_lines = apply_corrections(doc.content, corrections)
        metadata = doc.metadata.copy() if doc.metadata else {}
        proof_reading = {
            "model": self.model,
            "corrections_count": len(corrected_lines),
            "corrected_lines": sorted(corrected_lines),
            "metadata_only": self.add_metadata_only,
            "corrections": [
                {"line_number": c.line_number, "corrected": c.corrected}
                for c in corrections
            ],
        }
        if self.include_diffs:
            diff_metadata = generate_all_diffs(doc.content, new_content)
            proof_reading.update(diff_metadata)

        metadata[self.NAME] = proof_reading
        final_content = new_content if not self.add_metadata_only else doc.content
        return Document(
            content=final_content,
            images=doc.images,
            title=doc.title,
            author=doc.author,
            created=doc.created,
            modified=doc.modified,
            source_path=doc.source_path,
            mime_type=doc.mime_type,
            metadata=metadata,
        )


if __name__ == "__main__":
    import anyenv

    async def main():
        # Create a test document with OCR errors
        test_content = """\
OCR Test Document
This 1s a test document with some common OCR errors.
VVords are sometimes m1staken for other characters.
Spaces occasionally getremoved between words.
The letter 'l' is often confused with the number '1'.
Line endings may be incorrectlybroken.
Punctuation,marks can be misplaced , or incorrect.
Special @characters# might not be recognized properly.
numbers like 5678 can be misread as S67B."""

        doc = Document(
            content=test_content,
            title="Test OCR Document",
            source_path="test_document.txt",
            mime_type="text/plain",
        )
        proofreader = LLMProofReader(
            model=DEFAULT_PROOF_READER_MODEL,
            include_diffs=True,
        )

        print("Original content:")
        print("-" * 50)
        print(test_content)
        print("-" * 50)
        print("Processing with LLM proof reader...")
        corrected_doc = await proofreader.process(doc)
        print("\nCorrected content:")
        print("-" * 50)
        print(corrected_doc.content)
        print("-" * 50)
        proof_reading = corrected_doc.metadata["proof_reading"]
        print("\nProof reading metadata:")
        for key, value in proof_reading.items():
            if key == "unified_diff":
                print(f"\n{key}:")
                print("-" * 50)
                print(value)
                print("-" * 50)
            elif key not in ("html_diff", "semantic_diff"):
                print(f"{key}: {value}")

            print(f"\nCorrected {proof_reading.get('corrections_count', 0)} lines")

        return corrected_doc

    corrected = anyenv.run_sync(main())
    print(f"\nProofreading complete! Processed document: {corrected.title}")



"""Converters configuration for the Streamlit app."""

from __future__ import annotations

from typing import TYPE_CHECKING

# from docler.converters.llm_provider import LLMConverter
# from docler.converters.markitdown_provider import MarkItDownConverter
from docler.converters.azure_provider import AzureConverter
from docler.converters.datalab_provider import DataLabConverter
from docler.converters.docling_provider import DoclingConverter
from docler.converters.llamaparse_provider import LlamaParseConverter
from docler.converters.marker_provider import MarkerConverter
from docler.converters.mistral_provider import MistralConverter
from docler.converters.upstage_provider import UpstageConverter


if TYPE_CHECKING:
    from docler.converters.base import DocumentConverter

ls: list[type[DocumentConverter]] = [
    DataLabConverter,
    DoclingConverter,
    MarkerConverter,
    MistralConverter,
    LlamaParseConverter,
    AzureConverter,
    # MarkItDownConverter
    UpstageConverter,
]
CONVERTERS = {cls.NAME: cls for cls in ls}



"""Document converter using Upstage's Document AI API."""

from __future__ import annotations

import base64
from collections import defaultdict
from typing import TYPE_CHECKING, Any, ClassVar

from mkdown import Document, Image, create_image_reference, create_page_break
import requests
from upathtools import to_upath

from docler.configs.converter_configs import UpstageConfig
from docler.converters.base import DocumentConverter
from docler.pdf_utils import extract_pdf_pages
from docler.utils import get_api_key


if TYPE_CHECKING:
    from schemez import MimeType

    from docler.common_types import PageRangeString, StrPath, SupportedLanguage
    from docler.configs.converter_configs import UpstageCategory, UpstageOCRType


# API endpoints
DOCUMENT_PARSE_BASE_URL = "https://api.upstage.ai/v1/document-digitization"
DOCUMENT_PARSE_DEFAULT_MODEL = "document-parse"


# https://console.upstage.ai/api/document-digitization/document-parsing


class UpstageConverter(DocumentConverter[UpstageConfig]):
    """Document converter using Upstage's Document AI API."""

    Config = UpstageConfig

    NAME = "upstage"
    REQUIRED_PACKAGES: ClassVar = {"requests"}
    SUPPORTED_MIME_TYPES: ClassVar[set[str]] = {
        "application/pdf",
        "image/jpeg",
        "image/png",
        "image/tiff",
        "application/msword",
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
    }

    def __init__(
        self,
        languages: list[SupportedLanguage] | None = None,
        *,
        page_range: PageRangeString | None = None,
        api_key: str | None = None,
        base_url: str = DOCUMENT_PARSE_BASE_URL,
        model: str = DOCUMENT_PARSE_DEFAULT_MODEL,
        ocr: UpstageOCRType = "auto",
        chart_recognition: bool = True,
        align_orientation: bool = False,
        base64_categories: set[UpstageCategory] | None = None,
    ):
        """Initialize the Upstage converter.

        Args:
            languages: List of supported languages (currently unused by Upstage API)
            page_range: Page range(s) to extract, like "1-5,7-10" (1-based)
            api_key: Upstage API key (falls back to UPSTAGE_API_KEY env var)
            base_url: API endpoint URL
            model: Model name for document parsing
            ocr: OCR mode ('auto' or 'force')
            chart_recognition: Whether to convert charts to tables.
            align_orientation: Whether to automatically detect and correct the orientation
            base64_categories: Element categories to encode in base64
        """
        super().__init__(languages=languages, page_range=page_range)
        self.api_key = api_key or get_api_key("UPSTAGE_API_KEY")
        self.base_url = base_url
        self.model = model
        self.ocr = ocr
        self.base64_categories = base64_categories or {"figure", "chart"}
        self.chart_recognition = chart_recognition
        self.align_orientation = align_orientation

    @property
    def price_per_page(self) -> float:
        """Price per page in USD."""
        return 0.01

    def _convert_path_sync(self, file_path: StrPath, mime_type: MimeType) -> Document:
        """Convert a document using Upstage's Document AI API.

        Args:
            file_path: Path to the document file
            mime_type: MIME type of the file

        Returns:
            Converted document with extracted text, images, and page break markers.

        Raises:
            ValueError: If conversion fails or response is malformed.
        """
        path = to_upath(file_path)
        file_content = path.read_bytes()
        if self.page_range is not None:
            file_content = extract_pdf_pages(file_content, self.page_range)
        headers = {"Authorization": f"Bearer {self.api_key}"}
        files = {"document": (path.name, file_content, mime_type)}
        data = {
            "ocr": self.ocr,
            "model": self.model,
            "output_formats": "['markdown']",
            "base64_encoding": str(list(self.base64_categories)),
            "chart_recognition": self.chart_recognition,
            "align_orientation": self.align_orientation,
        }

        try:
            response = requests.post(
                self.base_url,
                headers=headers,
                files=files,
                data=data,
                timeout=300,
            )
            response.raise_for_status()
            result = response.json()
        except requests.HTTPError as e:
            msg = f"Upstage API error: {e.response.text if e.response else str(e)}"
            raise ValueError(msg) from e
        except Exception as e:
            msg = f"Failed to convert document via Upstage API: {e}"
            self.logger.exception(msg)
            raise ValueError(msg) from e

        content_data = result.get("content", {})
        initial_markdown = content_data.get("markdown")
        if not initial_markdown:
            msg = "No content found in Upstage API response."
            # self.logger.warning("Full Upstage response: %s", result)
            raise ValueError(msg)

        elements = result.get("elements", [])
        max_page = result.get("usage", {}).get("pages", 0)

        # Start with the page 1 marker
        first_page_marker = create_page_break(next_page=1, newline_separators=1).lstrip()
        modified_markdown = first_page_marker + initial_markdown.lstrip()

        if max_page > 1 and elements:
            elements_by_page: dict[int, list[dict[str, Any]]] = defaultdict(list)
            for element in elements:
                page_num = element.get("page")
                if page_num is not None:
                    elements_by_page[page_num].append(element)

            for page_num in elements_by_page:  # noqa: PLC0206
                elements_by_page[page_num].sort(key=lambda x: x.get("id", 0))

            insertion_offset = len(first_page_marker)  # Search after the page 1 marker
            for page_num in range(2, max_page + 1):
                if page_num not in elements_by_page or not elements_by_page[page_num]:
                    continue

                # Find the first non-empty markdown content for the anchor
                first_element_md = ""
                for elem in elements_by_page[page_num]:
                    first_element_md = elem.get("content", {}).get("markdown", "")
                    if first_element_md:
                        break
                if not first_element_md:
                    msg = "Could not find non-empty element md anchor for page %d"
                    self.logger.warning(msg, page_num)
                    continue

                # Find the position using the offset
                found_index = modified_markdown.find(first_element_md, insertion_offset)

                if found_index != -1:
                    marker = create_page_break(next_page=page_num, newline_separators=1)
                    modified_markdown = (
                        modified_markdown[:found_index]
                        + marker
                        + modified_markdown[found_index:]
                    )
                    insertion_offset = found_index + len(marker) + len(first_element_md)
                else:
                    msg = "Could not find insertion point for page break before page %d"
                    self.logger.warning(msg, page_num)

        images: list[Image] = []
        image_counter = 0
        # Use a temporary variable for markdown content during image replacement
        # to avoid interfering with the page break logic's offset calculations if
        # image placeholders were part of the first element's markdown.
        content_for_image_replacement = modified_markdown

        for element in elements:
            category = element.get("category")
            base64_data = element.get("base64_encoding")

            if category in self.base64_categories and base64_data:
                image_id = f"img-{image_counter}"
                image_counter += 1

                if base64_data.startswith("data:image/"):
                    mime_parts = base64_data.split(";")[0].split(":")
                    img_mime_type = mime_parts[1] if len(mime_parts) > 1 else "image/png"
                    img_data = base64_data.split(",", 1)[1]  # Get data after comma
                else:
                    img_mime_type = "image/png"
                    img_data = base64_data  # Assume it's pure base64

                img_bytes = base64.b64decode(img_data)
                ext = img_mime_type.split("/")[-1]
                # Handle potential complex mime types like 'svg+xml'
                ext = ext.split("+")[0]
                filename = f"{image_id}.{ext}"

                image = Image(
                    id=image_id,
                    content=img_bytes,
                    mime_type=img_mime_type,
                    filename=filename,
                )
                images.append(image)

                # Replace the *first* available placeholder in the markdown
                placeholder = "![image](/image/placeholder)"
                if placeholder in content_for_image_replacement:
                    img_ref = create_image_reference(image_id, filename)
                    content_for_image_replacement = content_for_image_replacement.replace(
                        placeholder, img_ref, 1
                    )
                else:
                    msg = "Found image data for %s but no placeholder left in markdown."
                    self.logger.warning(msg, image_id)
        modified_markdown = content_for_image_replacement
        return Document(
            content=modified_markdown.strip(),
            images=images,
            title=path.stem,
            source_path=str(path),
            mime_type=mime_type,
            metadata=result.get("metadata", {}),
        )


if __name__ == "__main__":
    import logging

    import anyenv

    logging.basicConfig(level=logging.INFO)  # Add basic logging for testing

    pdf_path = "src/docler/resources/pdf_sample.pdf"  # Adjust path if needed
    converter = UpstageConverter()
    result_doc = anyenv.run_sync(converter.convert_file(pdf_path))
    print("--- Converted Markdown ---")
    print(result_doc.content)
    print("\n--- Extracted Images ---")
    print(result_doc.images)
    print(f"\nPage Count: {result_doc.page_count}")



"""Aggregated document converter that delegates to appropriate converters."""

from __future__ import annotations

from typing import TYPE_CHECKING

import anyenv

from docler.configs.converter_configs import AggregatedConverterConfig
from docler.converters.base import DocumentConverter
from docler.converters.registry import ConverterRegistry


if TYPE_CHECKING:
    from mkdown import Document
    from schemez import MimeType

    from docler.common_types import PageRangeString, StrPath, SupportedLanguage


class AggregatedConverter(DocumentConverter[AggregatedConverterConfig]):
    """Converter that delegates to specialized converters based on mime type.

    This converter uses a registry of specialized converters and selects
    the most appropriate one for each file type. It acts as a single entry
    point that supports all file types supported by any registered converter.
    """

    Config = AggregatedConverterConfig
    NAME = "aggregated"

    def __init__(
        self,
        languages: list[SupportedLanguage] | None = None,
        *,
        page_range: PageRangeString | None = None,
        registry: ConverterRegistry | None = None,
    ):
        """Initialize the aggregated converter.

        Args:
            languages: Languages to use for conversion
            page_range: Page range to extract.
            registry: Existing registry to use, or create a new one if None
        """
        super().__init__(languages=languages, page_range=page_range)
        self._registry = registry or ConverterRegistry.create_default(languages=languages)

    @classmethod
    def from_config(cls, config: AggregatedConverterConfig) -> AggregatedConverter:
        """Create an AggregatedConverter instance from a configuration."""
        registry = ConverterRegistry()
        for converter_config in config.converters:
            converter = converter_config.get_provider()
            registry.register(converter)
        for mime_or_ext, converter_name in config.mime_preferences.items():
            registry.set_preference(mime_or_ext, converter_name)

        return cls(registry=registry)

    def to_config(self) -> AggregatedConverterConfig:
        """Extract configuration from the converter instance."""
        config = AggregatedConverterConfig()
        for converter in self._registry._converters:
            config.converters.append(converter.to_config())
        config.mime_preferences = dict(self._registry._preferences)
        return config

    def get_supported_mime_types(self) -> set[str]:
        """Get all MIME types supported by registered converters."""
        return self._registry.get_supported_mime_types()

    def set_converter_preference(self, mime_or_extension: str, converter_name: str):
        """Set a preference for which converter to use for a specific file type.

        Args:
            mime_or_extension: MIME type ('application/pdf') or file extension ('.pdf')
            converter_name: Name of the preferred converter ('mistral')
        """
        self._registry.set_preference(mime_or_extension, converter_name)

    async def _convert_path_async(
        self,
        file_path: StrPath,
        mime_type: MimeType,
    ) -> Document:
        """Delegate conversion to the appropriate converter."""
        converter = self._registry.get_converter(str(file_path), mime_type)
        if not converter:
            msg = f"No converter found for file: {file_path} (mime type: {mime_type})"
            raise ValueError(msg)

        try:
            return await converter._convert_path_async(file_path, mime_type)
        except NotImplementedError:
            return await anyenv.run_in_thread(
                converter._convert_path_sync, file_path, mime_type
            )



"""Logging configuration for docler."""

from __future__ import annotations

import logging


def get_logger(name: str) -> logging.Logger:
    """Get a logger for the given name."""
    return logging.getLogger(f"docler.{name}")



"""Pinecone Vector Store manager with asyncio support."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Literal, cast

from pydantic import SecretStr

from docler.configs.vector_db_configs import PineconeConfig
from docler.utils import get_api_key
from docler.vector_db.base import BaseVectorDB
from docler.vector_db.base_manager import VectorManagerBase
from docler.vector_db.dbs.pinecone_db.db import PineconeBackend
from docler.vector_db.dbs.pinecone_db.utils import to_vector_store_info


if TYPE_CHECKING:
    from docler.configs.vector_db_configs import PineconeCloud, PineconeRegion
    from docler.models import VectorStoreInfo


Metric = Literal["cosine", "euclidean", "dotproduct"]


class PineconeVectorManager(VectorManagerBase[PineconeConfig]):
    """Manager for Pinecone Vector Stores with asyncio support."""

    Config = PineconeConfig
    NAME = "pinecone"
    REQUIRED_PACKAGES: ClassVar = {"pinecone-client"}

    def __init__(self, api_key: str | None = None):
        """Initialize the Pinecone Vector Store manager."""
        super().__init__()
        self.api_key = api_key or get_api_key("PINECONE_API_KEY")

    @classmethod
    def from_config(cls, config: PineconeConfig) -> PineconeVectorManager:
        """Create instance from configuration."""
        key = config.api_key.get_secret_value() if config.api_key else None
        return cls(api_key=key)

    def to_config(self) -> PineconeConfig:
        """Extract configuration from instance."""
        return PineconeConfig(api_key=SecretStr(self.api_key) if self.api_key else None)

    @property
    def name(self) -> str:
        """Name of this vector database provider."""
        return self.NAME

    async def list_vector_stores(self) -> list[VectorStoreInfo]:
        """List all available vector stores for this provider."""
        from pinecone import PineconeAsyncio

        async with PineconeAsyncio(api_key=self.api_key) as client:
            indexes = await client.list_indexes()
            return [to_vector_store_info(idx) for idx in indexes]

    async def create_vector_store(
        self,
        name: str,
        dimension: int = 1536,
        metric: Metric = "cosine",
        cloud: PineconeCloud = "aws",
        region: PineconeRegion = "us-east-1",
        namespace: str = "default",
        **kwargs,
    ) -> BaseVectorDB:
        """Create a new vector store."""
        from pinecone import PineconeAsyncio, ServerlessSpec

        if await self.has_vector_store(name):
            msg = f"Index {name!r} already exists"
            raise ValueError(msg)
        async with PineconeAsyncio(api_key=self.api_key) as client:
            spec = ServerlessSpec(cloud=cloud.lower(), region=region)
            await client.create_index(name, spec, dimension=dimension, metric=metric)
            index_info = await client.describe_index(name)
            db = PineconeBackend(
                api_key=self.api_key,
                host=index_info.host,
                dimension=dimension,
                namespace=namespace,
            )
            return cast(BaseVectorDB, db)  # type: ignore[override]

    async def get_vector_store(self, name: str, **kwargs) -> BaseVectorDB:
        """Get a connection to an existing vector store."""
        from pinecone import PineconeAsyncio

        try:
            if not await self.has_vector_store(name):
                msg = f"Index {name} does not exist"
                raise ValueError(msg)  # noqa: TRY301
            async with PineconeAsyncio(api_key=self.api_key) as client:
                index_info = await client.describe_index(name)
            db = PineconeBackend(
                api_key=self.api_key,
                host=index_info.host,
                dimension=index_info.dimension,
                namespace=kwargs.get("namespace", "default"),
            )

        except Exception as e:
            msg = f"Failed to connect to vector store {name}: {e}"
            self.logger.exception(msg)
            raise ValueError(msg) from e
        else:
            return cast(BaseVectorDB, db)

    async def delete_vector_store(self, name: str) -> bool:
        """Delete a vector store."""
        from pinecone import PineconeAsyncio

        try:
            if not await self.has_vector_store(name):
                return False
            async with PineconeAsyncio(api_key=self.api_key) as client:
                index_info = await client.describe_index(name)
                if index_info.deletion_protection == "enabled":
                    await client.configure_index(name, deletion_protection="disabled")
                await client.delete_index(name)

        except Exception:
            self.logger.exception("Error deleting vector store %s", name)
            return False
        else:
            return True

    async def close(self):
        """Close all vector store connections."""


if __name__ == "__main__":
    import anyenv

    async def main():
        manager = PineconeVectorManager()
        indexes = await manager.list_vector_stores()
        print(indexes)
        await manager.close()

    anyenv.run_sync(main())



"""ChromaDB vector store manager implementation."""

from __future__ import annotations

from pathlib import Path
import shutil
from typing import TYPE_CHECKING, ClassVar, cast

from docler.configs.vector_db_configs import ChromaConfig
from docler.models import VectorStoreInfo
from docler.process_runner import ProcessRunner
from docler.vector_db.base import BaseVectorDB
from docler.vector_db.base_manager import VectorManagerBase
from docler.vector_db.dbs.chroma_db.db import ChromaBackend


if TYPE_CHECKING:
    import chromadb

    from docler.common_types import StrPath


class ChromaVectorManager(VectorManagerBase[ChromaConfig]):
    """Manager for ChromaDB vector stores with fully async implementation."""

    Config = ChromaConfig
    NAME = "chroma"
    REQUIRED_PACKAGES: ClassVar = {"chromadb"}

    def __init__(
        self,
        *,
        persist_directory: str | None = None,
        host: str | None = None,
        port: int = 8000,
        ssl: bool = False,
        headers: dict[str, str] | None = None,
    ):
        """Initialize the ChromaDB vector store manager.

        Args:
            persist_directory: Directory for persistent storage
            host: Hostname for remote ChromaDB server
            port: Port for remote ChromaDB server
            ssl: Whether to use SSL for server connection
            headers: Optional headers for server connection
        """
        super().__init__()
        self.persist_directory = persist_directory
        self.host = host or "localhost"
        self.port = port
        self.ssl = ssl
        self.headers = headers
        self._list_client: chromadb.AsyncClientAPI | None = None

    @staticmethod
    def run_server(path: StrPath, **kwargs) -> ProcessRunner:
        cmd = f"chroma run {path}"
        # if self.port:
        #     cmd += f" --port {self.port}"
        return ProcessRunner(cmd, **kwargs)

    @property
    def name(self) -> str:
        """Name of this vector database provider."""
        return self.NAME

    @classmethod
    def from_config(cls, config: ChromaConfig) -> ChromaVectorManager:
        """Create instance from configuration."""
        return cls(persist_directory=config.persist_directory)

    def to_config(self) -> ChromaConfig:
        """Extract configuration from instance."""
        return ChromaConfig(persist_directory=self.persist_directory)

    async def _get_list_client(self) -> chromadb.AsyncClientAPI:
        """Get a client for listing collections."""
        import chromadb

        if not self._list_client:
            self._list_client = await chromadb.AsyncHttpClient(
                host=self.host,
                port=self.port,
                ssl=self.ssl,
                headers=self.headers,
            )
        return self._list_client

    async def list_vector_stores(self) -> list[VectorStoreInfo]:
        """List all available vector stores (collections) for this provider."""
        try:
            client = await self._get_list_client()
            collections = await client.list_collections()
            return [VectorStoreInfo(db_id=c.name, name=c.name) for c in collections]
        except Exception:
            self.logger.exception("Error listing ChromaDB collections")
            return []

    async def create_vector_store(self, name: str, **kwargs) -> BaseVectorDB:
        """Create a new vector store (collection)."""
        try:
            db = ChromaBackend(
                vector_store_id=name,
                persist_directory=self.persist_directory,
                **kwargs,
            )
            return cast(BaseVectorDB, db)

        except Exception as e:
            msg = f"Failed to create ChromaDB collection: {e}"
            self.logger.exception(msg)
            raise ValueError(msg) from e

    async def get_vector_store(
        self,
        name: str,
        **kwargs,
    ) -> BaseVectorDB:
        """Get a connection to an existing collection."""
        try:
            client = await self._get_list_client()
            collection_names = await client.list_collections()
            if name not in collection_names:
                msg = f"Collection {name!r} not found in ChromaDB"
                raise ValueError(msg)  # noqa: TRY301
            db = ChromaBackend(
                vector_store_id=name,
                persist_directory=self.persist_directory,
                **kwargs,
            )

            return cast(BaseVectorDB, db)

        except Exception as e:
            msg = f"Failed to connect to ChromaDB collection {name}: {e}"
            self.logger.exception(msg)
            raise ValueError(msg) from e

    async def delete_vector_store(self, name: str) -> bool:
        """Delete a vector store (collection)."""
        try:
            client = await self._get_list_client()
            await client.delete_collection(name=name)
            if self.persist_directory:
                collection_path = Path(self.persist_directory) / name
                if collection_path.exists():
                    shutil.rmtree(collection_path, ignore_errors=True)

        except Exception:
            self.logger.exception("Failed to delete collection %s", name)
            return False
        else:
            return True

    async def close(self):
        """Close all vector store connections."""
        if self._list_client:
            try:
                await self._list_client.reset()
            except Exception as e:  # noqa: BLE001
                self.logger.warning("Error closing list client: %s", e)

            self._list_client = None


if __name__ == "__main__":
    import asyncio
    import logging

    logging.basicConfig(level=logging.INFO)

    async def main():
        async with ChromaVectorManager.run_server(
            "test",
            wait_output=[".*Uvicorn running.*"],
            wait_timeout=10,
        ):
            manager = ChromaVectorManager()
            await manager.create_vector_store("test")
            await manager.list_vector_stores()
            await manager.delete_vector_store("test")
            await manager.close()

    asyncio.run(main())



from __future__ import annotations

from docler.vector_db.dbs.qdrant_db.db import QdrantBackend
from docler.vector_db.dbs.qdrant_db.manager import QdrantVectorManager

__all__ = ["QdrantBackend", "QdrantVectorManager"]



"""AI-powered document and chunk metadata annotation."""

from __future__ import annotations

from itertools import batched
from typing import TYPE_CHECKING, ClassVar, TypeVar

import anyenv
from schemez import Schema

from docler.annotators.base import Annotator
from docler.common_types import DEFAULT_ANNOTATOR_MODEL
from docler.configs.annotator_configs import (
    DEFAULT_DOC_PROMPT_TEMPLATE,
    DEFAULT_DOC_SYSTEM_PROMPT,
    AIDocumentAnnotatorConfig,
)


if TYPE_CHECKING:
    from llmling_agent import StructuredAgent

    from docler.models import ChunkedDocument


class DefaultMetadata(Schema):
    """Default metadata for a document or chunk."""

    topics: list[str]
    """Topics/categories."""

    keywords: list[str]
    """Keywords."""

    entities: list[str]
    """Main entities."""


T = TypeVar("T", bound=Schema)


class AIDocumentAnnotator[TMetadata](Annotator[AIDocumentAnnotatorConfig]):
    """AI-based document and chunk annotator.

    Enhances documents and chunks with metadata.

    Type Parameters:
        T: Type of metadata model to use. Must be a schemez Schema.
    """

    Config = AIDocumentAnnotatorConfig
    REQUIRED_PACKAGES: ClassVar = {"llmling-agent"}

    def __init__(
        self,
        model: str | None = None,
        system_prompt: str | None = None,
        user_prompt: str | None = None,
        metadata_model: type[TMetadata] = DefaultMetadata,  # type: ignore
        max_context_length: int = 1500,
        batch_size: int = 5,
    ):
        """Initialize the AI document annotator.

        Args:
            model: LLM model to use for annotation
            system_prompt: Optional custom prompt for annotation
            user_prompt: Optional custom prompt for annotation
            metadata_model: Pydantic model class for metadata structure
            max_context_length: Maximum length of context for annotation
            batch_size: Number of chunks to process in parallel
        """
        super().__init__()
        self.model = model or DEFAULT_ANNOTATOR_MODEL
        self.system_prompt = system_prompt or DEFAULT_DOC_SYSTEM_PROMPT
        self.user_prompt = user_prompt or DEFAULT_DOC_PROMPT_TEMPLATE
        self.metadata_model = metadata_model
        self.max_context_length = max_context_length
        self.batch_size = batch_size

    async def annotate(self, document: ChunkedDocument) -> ChunkedDocument:
        """Annotate document and chunks with AI-generated metadata.

        Args:
            document: Chunked document to annotate

        Returns:
            Document with enhanced metadata
        """
        from llmling_agent import Agent

        agent: StructuredAgent[None, TMetadata] = Agent[None](  # type: ignore
            model=self.model,
            system_prompt=self.system_prompt,
        ).to_structured(self.metadata_model)

        # Get a condensed version of the document for context
        context = (
            document.content[: self.max_context_length] + "..."
            if self.max_context_length and len(document.content) > self.max_context_length
            else document.content
        )

        # Process chunks in batches
        for batch in batched(document.chunks, self.batch_size, strict=False):
            tasks = []

            for chunk in batch:
                prompt = self.user_prompt.format(context=context, chunk=chunk.content)
                tasks.append(agent.run(prompt))

            try:
                results = await anyenv.gather(*tasks)
                for chunk, result in zip(batch, results):
                    metadata = result.content.model_dump()  # type: ignore
                    chunk.metadata |= metadata
            except Exception:
                self.logger.exception("Error annotating batch")

        return document


if __name__ == "__main__":
    import asyncio

    from docler.models import ChunkedDocument

    async def main():
        text = "Test"
        annotator = AIDocumentAnnotator[DefaultMetadata]()
        chunked_doc = ChunkedDocument(content=text)
        result = await annotator.annotate(chunked_doc)
        print(result)

    asyncio.run(main())



"""Base markdown chunking implementation."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mkdown import TextChunk

from docler.chunkers.base import TextChunker
from docler.chunkers.markdown_chunker.utils import assign_images, split_by_headers
from docler.configs.chunker_configs import MarkdownChunkerConfig


if TYPE_CHECKING:
    from collections.abc import Iterator

    from mkdown import Document, Image


class MarkdownChunker(TextChunker[MarkdownChunkerConfig]):
    """Header-based markdown chunker with fallback to size-based chunks."""

    Config = MarkdownChunkerConfig
    NAME = "markdown"

    def __init__(
        self,
        *,
        min_chunk_size: int = 200,
        max_chunk_size: int = 1500,
        chunk_overlap: int = 50,
    ):
        """Initialize chunker.

        Args:
            min_chunk_size: Minimum characters per chunk
            max_chunk_size: Maximum characters per chunk
            chunk_overlap: Character overlap between chunks
        """
        self.min_chunk_size = min_chunk_size
        self.max_chunk_size = max_chunk_size
        self.chunk_overlap = chunk_overlap

    def _fallback_split(
        self, content: str, images: list[Image]
    ) -> Iterator[tuple[str, list[Image]]]:
        """Split content by size when no headers exist."""
        start = 0
        while start < len(content):
            chunk_content = content[start : start + self.max_chunk_size]
            chunk_content, chunk_images = assign_images(chunk_content, images)
            yield chunk_content, chunk_images
            start += self.max_chunk_size - self.chunk_overlap

    async def split(
        self,
        doc: Document,
        extra_metadata: dict[str, Any] | None = None,
    ) -> list[TextChunk]:
        """Split document into chunks."""
        chunks: list[TextChunk] = []
        chunk_idx = 0

        # Try header-based splitting first
        header_sections = list(split_by_headers(doc.content))
        if not header_sections:
            # Fallback to size-based if no headers
            for content, images in self._fallback_split(doc.content, doc.images):
                chunk = TextChunk(
                    content=content,
                    source_doc_id=doc.source_path or "",
                    chunk_index=chunk_idx,
                    images=images,
                    metadata=extra_metadata or {},
                )
                chunks.append(chunk)
                chunk_idx += 1
            return chunks

        # Process header sections
        for header, content, level in header_sections:
            meta = {**(extra_metadata or {}), "header": header, "level": level}
            if len(content) > self.max_chunk_size:
                for sub_content, images in self._fallback_split(content, doc.images):
                    chunk = TextChunk(
                        content=f"{header}\n\n{sub_content}",
                        source_doc_id=doc.source_path or "",
                        chunk_index=chunk_idx,
                        images=images,
                        metadata=meta,
                    )
                    chunks.append(chunk)
                    chunk_idx += 1
            else:
                content, images = assign_images(content, doc.images)
                chunk = TextChunk(
                    content=f"{header}\n\n{content}",
                    source_doc_id=doc.source_path or "",
                    chunk_index=chunk_idx,
                    images=images,
                    metadata=meta,
                )
                chunks.append(chunk)
                chunk_idx += 1

        return chunks



"""Process runner with configurable wait conditions."""

from __future__ import annotations

import asyncio
import contextlib
import http.client
import os
import re
import socket
import subprocess
import threading
from typing import IO, TYPE_CHECKING, Literal, Self
import urllib.parse
import weakref


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

RegexPattern = str

_process_registry: dict[tuple[str, ...], weakref.ref[subprocess.Popen]] = {}
_registry_lock = asyncio.Lock()


def _clean_registry():
    """Remove dead processes from registry."""
    dead_keys = []
    for key, ref in _process_registry.items():
        process = ref()
        if process is None or process.poll() is not None:
            dead_keys.append(key)

    for key in dead_keys:
        del _process_registry[key]


async def _check_http(url: str) -> bool:
    parsed = urllib.parse.urlparse(url)
    conn = http.client.HTTPConnection(parsed.netloc)
    try:
        conn.request("GET", parsed.path or "/")
        response = conn.getresponse()
    except (ConnectionRefusedError, socket.gaierror):
        return False
    else:
        return 200 <= response.status < 400  # noqa: PLR2004
    finally:
        conn.close()


async def _check_tcp(host: str, port: int) -> bool:
    try:
        _, writer = await asyncio.open_connection(host, port)
        writer.close()
        await writer.wait_closed()
    except (ConnectionRefusedError, socket.gaierror):
        return False
    else:
        return True


async def _check_predicate(
    pred: Callable[[], bool] | Callable[[], Awaitable[bool]],
) -> bool:
    if asyncio.iscoroutinefunction(pred):
        return await pred()
    return await asyncio.to_thread(pred)  # type: ignore


def _reader_thread(
    stream: IO[bytes],
    buffer: list[str],
    patterns: dict[re.Pattern[str], bool] | None = None,
):
    """Thread to read from process streams."""
    try:
        for line in iter(stream.readline, b""):
            try:
                decoded = line.decode(errors="ignore").rstrip()
                buffer.append(decoded)  # Store the line in the buffer
                if patterns:
                    for pattern in patterns:
                        if pattern.search(decoded):
                            patterns[pattern] = True
            except Exception:  # noqa: BLE001
                continue  # Skip any decoding errors
    except ValueError:
        # Stream likely closed
        pass


class ProcessRunner:
    def __init__(
        self,
        command: list[str] | str,
        *,
        reuse: bool = False,
        wait_http: list[str] | None = None,  # ["http://localhost:8000/health"]
        wait_tcp: list[tuple[str, int]] | None = None,  # [("localhost", 6379)]
        wait_predicates: list[Callable[[], bool] | Callable[[], Awaitable[bool]]]
        | None = None,
        wait_output: list[RegexPattern] | None = None,
        wait_stderr: list[RegexPattern] | None = None,
        wait_timeout: float = 30.0,
        poll_interval: float = 0.1,
        cleanup_timeout: float = 5.0,
    ):
        self.cleanup_timeout = cleanup_timeout
        self.command = command if isinstance(command, list) else command.split()
        self.command_key = tuple(self.command)
        self.reuse = reuse
        self.wait_http = wait_http or []
        self.wait_tcp = wait_tcp or []
        self.wait_predicates = wait_predicates or []
        self.wait_output = [re.compile(p) for p in (wait_output or [])]
        self.wait_stderr = [re.compile(p) for p in (wait_stderr or [])]
        self.wait_timeout = wait_timeout
        self.poll_interval = poll_interval
        self.process: subprocess.Popen | None = None
        self._stdout_patterns_found = dict.fromkeys(self.wait_output, False)
        self._stderr_patterns_found = dict.fromkeys(self.wait_stderr, False)
        self._stdout_buffer: list[str] = []
        self._stderr_buffer: list[str] = []
        # Threads for stream reading
        self._stdout_reader: threading.Thread | None = None
        self._stderr_reader: threading.Thread | None = None

    async def _wait_for_conditions(self):
        async def check_all() -> bool:
            # Check HTTP endpoints
            http_results = await asyncio.gather(
                *(_check_http(url) for url in self.wait_http)
            )
            if not all(http_results):
                return False

            # Check TCP ports
            tcp_results = await asyncio.gather(
                *(_check_tcp(h, p) for h, p in self.wait_tcp)
            )
            if not all(tcp_results):
                return False

            # Check predicates
            pred_results = await asyncio.gather(
                *(_check_predicate(p) for p in self.wait_predicates)
            )
            if not all(pred_results):
                return False

            # Check output patterns
            if not all(self._stdout_patterns_found.values()):
                return False
            return all(self._stderr_patterns_found.values())

        start_time = asyncio.get_event_loop().time()
        while True:
            if await check_all():
                return

            if (asyncio.get_event_loop().time() - start_time) > self.wait_timeout:
                msg = "Timeout waiting for conditions"
                raise TimeoutError(msg)

            await asyncio.sleep(self.poll_interval)

    async def __aenter__(self) -> Self:
        if self.reuse:
            async with _registry_lock:
                _clean_registry()
                if self.command_key in _process_registry:
                    existing_process = _process_registry[self.command_key]()
                    if existing_process is not None and existing_process.poll() is None:
                        self.process = existing_process
                        return self

        # Use shell=True for Windows and non-standard command execution
        is_windows = os.name == "nt"
        # For Windows, use the raw command string with shell=True to avoid issues
        if is_windows and isinstance(self.command, list):
            cmd_str = subprocess.list2cmdline(self.command)
        else:
            cmd_str = (
                self.command
                if isinstance(self.command, str)
                else subprocess.list2cmdline(self.command)
            )

        self.process = subprocess.Popen(
            cmd_str,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=is_windows,  # Use shell on Windows
            text=False,  # Keep as bytes for proper line handling
            bufsize=0,  # Unbuffered
        )

        _process_registry[self.command_key] = weakref.ref(self.process)

        # Start reader threads
        self._stdout_reader = threading.Thread(
            target=_reader_thread,
            args=(self.process.stdout, self._stdout_buffer, self._stdout_patterns_found),
            daemon=True,
        )
        self._stdout_reader.start()

        self._stderr_reader = threading.Thread(
            target=_reader_thread,
            args=(self.process.stderr, self._stderr_buffer, self._stderr_patterns_found),
            daemon=True,
        )
        self._stderr_reader.start()

        has_conditions = (
            self.wait_http
            or self.wait_tcp
            or self.wait_predicates
            or self.wait_output
            or self.wait_stderr
        )

        if has_conditions:
            try:
                await self._wait_for_conditions()
            except Exception as e:
                self.process.kill()
                msg = "Failed waiting for process to be ready"
                raise RuntimeError(msg) from e

        return self

    async def __aexit__(self, *_):
        """Clean up process with timeout."""
        if self.process is None:
            return
        try:
            self.process.kill()
            with contextlib.suppress(subprocess.TimeoutExpired):
                self.process.wait(timeout=self.cleanup_timeout)
        except ProcessLookupError:
            pass

    @property
    def pid(self) -> int | None:
        """Return process ID if running."""
        return self.process.pid if self.process else None

    @property
    def returncode(self) -> int | None:
        """Return exit code if process has finished."""
        return self.process.poll() if self.process else None

    def send_signal(self, sig: int):
        """Send a signal to the process."""
        if self.process:
            self.process.send_signal(sig)

    async def get_output(self, stream: Literal["stdout", "stderr"] = "stdout") -> str:
        """Get current output content from stdout or stderr."""
        buffer = self._stderr_buffer if stream == "stderr" else self._stdout_buffer
        return "\n".join(buffer)


if __name__ == "__main__":

    async def main():
        docker_cmd = "docker run -p 6333:6333 -p 6334:6334 qdrant/qdrant"
        print(f"Running command: {docker_cmd}")
        async with ProcessRunner(docker_cmd, wait_tcp=[("localhost", 6333)]) as runner:
            print(f"Process running with PID {runner.pid}")
            print("\nSTDOUT:")
            print(await runner.get_output())
            print("\nSTDERR:")
            print(await runner.get_output("stderr"))

        print("Process completed")

    asyncio.run(main())



"""Pinecone vector store backend implementation."""

from __future__ import annotations

import base64
from typing import TYPE_CHECKING, Any

from docler.models import SearchResult, VectorStoreInfo


if TYPE_CHECKING:
    from pinecone import IndexModel
    from pinecone.core.openapi.db_data.models import ScoredVector


def convert_filters(filters: dict[str, Any]) -> dict:
    """Convert standard filters to Pinecone filter format."""
    pinecone_filter = {}
    for key, value in filters.items():
        pinecone_filter[key] = {"$in": value} if isinstance(value, list) else value
    return pinecone_filter


def prepare_metadata(metadata: dict[str, Any]) -> dict[str, Any]:
    """Prepare metadata for Pinecone storage."""
    import anyenv

    prepared = {}
    for key, value in metadata.items():
        if isinstance(value, str | int | float | bool):
            prepared[key] = value
        elif isinstance(value, list | dict):
            # Convert complex types to JSON strings
            try:
                # First try to store it directly if simple enough
                if isinstance(value, list) and all(
                    isinstance(x, str | int | float | bool) for x in value
                ):
                    prepared[key] = value  # type: ignore
                else:
                    prepared[f"{key}_json"] = anyenv.dump_json(value)
            except (TypeError, ValueError):
                dumped = anyenv.dump_json(str(value)).encode()
                prepared[f"{key}_b64"] = base64.b64encode(dumped).decode()
        elif value is not None:
            prepared[key] = str(value)

    # Ensure text field is preserved
    if "text" in metadata:
        prepared["text"] = str(metadata["text"])

    return prepared


def restore_metadata(metadata: dict[str, Any]) -> dict[str, Any]:
    """Restore encoded metadata fields."""
    import anyenv

    restored = metadata.copy()

    for key in list(restored.keys()):
        if key.endswith("_json") and key[:-5] not in restored:
            try:
                base_key = key[:-5]
                restored[base_key] = anyenv.load_json(restored[key])
                del restored[key]
            except Exception:  # noqa: BLE001
                pass

        elif key.endswith("_b64") and key[:-4] not in restored:
            try:
                base_key = key[:-4]
                json_str = base64.b64decode(restored[key]).decode()
                restored[base_key] = anyenv.load_json(json_str)
                del restored[key]
            except Exception:  # noqa: BLE001
                pass

    return restored


def to_search_result(match: ScoredVector) -> SearchResult:
    """Convert a Pinecone match to a SearchResult."""
    raw_metadata = match.metadata or {}
    metadata = restore_metadata(raw_metadata)
    text = metadata.pop("text", None) if isinstance(metadata, dict) else None
    score = match.score or 0.0
    return SearchResult(chunk_id=match.id, score=score, metadata=metadata, text=text)


def to_vector_store_info(index: IndexModel) -> VectorStoreInfo:
    """Convert a Pinecone index to a VectorStoreInfo."""
    meta = dict(
        dimension=index.dimension,
        metric=index.metric,
        status=index.status.state if index.status else None,
        ready=index.status.ready if index.status else False,
        vector_type=index.vector_type,
        tags=index.tags,
    )
    return VectorStoreInfo(db_id=index.host, name=index.name, metadata=meta)



"""Language code handling for OCR backends."""

from __future__ import annotations

from typing import TYPE_CHECKING

from docler.common_types import (
    MAC_CODES,
    RAPID_CODES,
    TESSERACT_CODES,
)


if TYPE_CHECKING:
    from docling.datamodel.pipeline_options import (
        EasyOcrOptions,
        OcrMacOptions,
        RapidOcrOptions,
        TesseractCliOcrOptions,
        TesseractOcrOptions,
    )

    from docler.common_types import PageRangeString, SupportedLanguage


def convert_languages(
    languages: list[SupportedLanguage],
    backend_type: type[
        EasyOcrOptions
        | TesseractCliOcrOptions
        | TesseractOcrOptions
        | OcrMacOptions
        | RapidOcrOptions
    ],
) -> list[str]:
    """Convert language codes for specific backend.

    Args:
        languages: List of language codes to convert
        backend_type: OCR backend class to convert for

    Returns:
        List of language codes in the format expected by the backend
    """
    from docling.datamodel.pipeline_options import (
        OcrMacOptions,
        RapidOcrOptions,
        TesseractCliOcrOptions,
        TesseractOcrOptions,
    )

    if backend_type in (TesseractCliOcrOptions, TesseractOcrOptions):
        return [TESSERACT_CODES[lang] for lang in languages]
    if backend_type == OcrMacOptions:
        return [MAC_CODES[lang] for lang in languages]
    if backend_type == RapidOcrOptions:
        return [RAPID_CODES[lang] for lang in languages]
    # EasyOCR uses standard 2-letter codes
    return list(languages)


def parse_page_range(page_range: PageRangeString) -> tuple[int, int]:
    """Convert a page range string to a tuple of (start, end) page numbers.

    Args:
        page_range: String like "1-5" or None. 1-based page numbers.

    Returns:
        Tuple of (start, end) page numbers (0-based) or None if no range specified.

    Raises:
        ValueError: If the page range format is invalid.
    """
    try:
        # Handle only first range for now (ignore possible additional ranges after comma)
        first_range = page_range.split(",")[0]
        start, end = map(int, first_range.split("-"))
        # Convert to 0-based indexing
        return (start - 1, end - 1)
    except ValueError as e:
        msg = f"Invalid page range format: {page_range}. Expected format: '1-5'"
        raise ValueError(msg) from e



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any

from mkdown import Image, create_image_reference, create_page_break


if TYPE_CHECKING:
    from collections.abc import Iterator

    from azure.ai.documentintelligence.models import AnalyzeResult


def to_image(response: Iterator[bytes], i: int) -> Image:
    content = b"".join(response)
    image_id = f"img-{i}"
    filename = f"{image_id}.png"
    return Image(id=image_id, content=content, mime_type="image/png", filename=filename)


def update_content(content: str, images: list[Image]) -> str:
    figure_pattern = r"<figure>(.*?)</figure>"
    figure_blocks = re.findall(figure_pattern, content, re.DOTALL)
    for i, block in enumerate(figure_blocks):
        if i < len(images):
            image = images[i]
            img_ref = create_image_reference(image.id, image.filename or "")
            content = content.replace(f"<figure>{block}</figure>", img_ref, 1)
    return content


def get_metadata(result: AnalyzeResult) -> dict[str, Any]:
    metadata = {}
    if result.documents:
        doc = result.documents[0]  # Get first document
        if doc.fields:
            metadata = {
                name: field.get("valueString") or field.get("content", "")
                for name, field in doc.fields.items()
            }
    return metadata


def replace_page_breaks(content: str) -> str:
    azure_marker = r"<!--\s*PageBreak\s*-->"
    page_num = 1

    def replace_marker(match: re.Match[str]) -> str:
        nonlocal page_num
        page_num += 1
        return create_page_break(next_page=page_num, newline_separators=1)

    processed_content = re.sub(azure_marker, replace_marker, content)
    first_page_marker = create_page_break(next_page=1, newline_separators=1).lstrip()
    return first_page_marker + processed_content



"""Document converter using LlamaParse."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

import anyenv
from mkdown import Document
from upathtools import to_upath

from docler.configs.converter_configs import LlamaParseConfig
from docler.converters.base import DocumentConverter
from docler.converters.llamaparse_provider.utils import process_response
from docler.log import get_logger
from docler.utils import get_api_key


if TYPE_CHECKING:
    from schemez import MimeType

    from docler.common_types import PageRangeString, StrPath, SupportedLanguage
    from docler.configs.converter_configs import LlamaParseMode


logger = get_logger(__name__)


class LlamaParseConverter(DocumentConverter[LlamaParseConfig]):
    """Document converter using LlamaParse."""

    Config = LlamaParseConfig

    NAME = "llamaparse"
    REQUIRED_PACKAGES: ClassVar = {"llama-parse"}
    SUPPORTED_MIME_TYPES: ClassVar[set[str]] = {
        # PDF
        "application/pdf",
        # Office Documents
        "application/msword",  # .doc
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document",  # .docx  # noqa: E501
        "application/vnd.ms-word.document.macroEnabled.12",  # .docm
        "application/vnd.ms-powerpoint",  # .ppt
        "application/vnd.openxmlformats-officedocument.presentationml.presentation",  # .pptx # noqa: E501
        "application/vnd.ms-powerpoint.presentation.macroEnabled.12",  # .pptm
        "application/vnd.ms-excel",  # .xls
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",  # .xlsx
        "application/vnd.ms-excel.sheet.macroEnabled.12",  # .xlsm
        "application/vnd.ms-excel.sheet.binary.macroEnabled.12",  # .xlsb
        # Open/Libre Office
        "application/vnd.oasis.opendocument.text",  # .odt
        "application/vnd.oasis.opendocument.spreadsheet",  # .ods
        "application/vnd.oasis.opendocument.presentation",  # .odp
        # Text formats
        "text/html",
        "text/markdown",
        "text/plain",  # .txt
        "text/rtf",  # .rtf
        "text/csv",  # .csv
        "text/tab-separated-values",  # .tsv
        "application/xml",  # .xml
        "application/epub+zip",  # .epub
        # Images
        "image/jpeg",  # .jpg, .jpeg
        "image/png",  # .png
        "image/gif",  # .gif
        "image/bmp",  # .bmp
        "image/svg+xml",  # .svg
        "image/tiff",  # .tiff
        "image/webp",  # .webp
        # Audio
        "audio/mpeg",  # .mp3
        "audio/mp4",  # .mp4 audio
        "audio/wav",  # .wav
        "audio/webm",  # .webm audio
        "audio/m4a",  # .m4a
    }

    def __init__(
        self,
        languages: list[SupportedLanguage] | None = None,
        *,
        page_range: PageRangeString | None = None,
        api_key: str | None = None,
        adaptive_long_table: bool = True,
        parse_mode: LlamaParseMode = "parse_page_with_llm",
        skip_diagonal_text: bool = False,
        disable_ocr: bool = False,
        continuous_mode: bool = True,
        html_tables: bool = False,
    ):
        """Initialize the LlamaParse converter.

        Args:
            languages: List of supported languages
            page_range: Page range(s) to extract, like "1-5,7-10" (1-based)
            api_key: LlamaParse API key, defaults to LLAMAPARSE_API_KEY env var
            adaptive_long_table: Whether to use adaptive long table
            parse_mode: Parse mode, defaults to "parse_page_with_llm"
            skip_diagonal_text: Whether to skip diagonal text
            disable_ocr: Whether to disable OCR for images
            continuous_mode: Whether to use continuous mode
            html_tables: Whether to output HTML tables instead of markdown
        """
        super().__init__(languages=languages, page_range=page_range)
        self.api_key = api_key or get_api_key("LLAMAPARSE_API_KEY")
        self.language = self.languages[0] if self.languages else None
        self.adaptive_long_table = adaptive_long_table
        self.parse_mode = parse_mode
        self.skip_diagonal_text = skip_diagonal_text
        self.disable_ocr = disable_ocr
        # self.bounding_box = None
        self.continuous_mode = continuous_mode
        self.html_tables = html_tables

    @property
    def price_per_page(self) -> float:
        """Price per page in USD."""
        return 0.0045

    def _convert_path_sync(self, file_path: StrPath, mime_type: MimeType) -> Document:
        """Convert a document using LlamaParse."""
        from llama_parse import LlamaParse, ResultType

        path = to_upath(file_path)
        parser = LlamaParse(
            api_key=self.api_key,
            result_type=ResultType.MD,
            language=self.language,
            adaptive_long_table=self.adaptive_long_table,
            parse_mode=self.parse_mode,
            target_pages=self.page_range,
            skip_diagonal_text=self.skip_diagonal_text,
            disable_ocr=self.disable_ocr,
            continuous_mode=self.continuous_mode,
            output_tables_as_HTML=self.html_tables,
            # take_screenshot=True,
            # we are doing page separators manually right now
            # page_separator=r'<!-- docler:page_break {"next_page":{pageNumber}} -->',
        )
        result = parser.get_json_result(str(path))
        content_parts, images = process_response(result, self.api_key)
        return Document(
            content="\n\n".join(content_parts),
            images=images,
            title=path.stem,
            source_path=str(path),
            mime_type=mime_type,
        )


if __name__ == "__main__":
    pdf_path = "src/docler/resources/pdf_sample.pdf"
    converter = LlamaParseConverter()
    result = anyenv.run_sync(converter.convert_file(pdf_path))
    print(result.content)



"""Command line interface for Docler document converter."""

from __future__ import annotations

from enum import Enum
import os
import pathlib
import subprocess
import sys

import typer

from docler.log import get_logger


cli = typer.Typer(help="Docler document converter CLI", no_args_is_help=True)

logger = get_logger(__name__)

PACKAGE_DIR = pathlib.Path(__file__).parent.parent / "docler_streamlit"


class TransportType(str, Enum):
    """Transport protocol types for MCP server."""

    STDIO = "stdio"
    SSE = "sse"
    STREAMABLE_HTTP = "streamable-http"


@cli.command()
def flow():
    """Start the Streamlit web interface."""
    app_path = str(PACKAGE_DIR / "app.py")
    try:
        cmd = [sys.executable, "-m", "streamlit", "run", app_path]
        subprocess.run(cmd, env=os.environ.copy(), check=True)
    except subprocess.CalledProcessError as e:
        # msg = f"Failed to start Streamlit: {e}"
        raise typer.Exit(1) from e
    except KeyboardInterrupt:
        logger.info("Shutting down...")


@cli.command("chunk_ui")
def chunk_ui():
    """Start the Streamlit web interface."""
    app_path = str(PACKAGE_DIR / "chunk_app.py")
    try:
        cmd = [sys.executable, "-m", "streamlit", "run", app_path]
        subprocess.run(cmd, env=os.environ.copy(), check=True)
    except subprocess.CalledProcessError as e:
        # msg = f"Failed to start Streamlit: {e}"
        raise typer.Exit(1) from e
    except KeyboardInterrupt:
        logger.info("Shutting down...")


@cli.command("ocr_ui")
def ocr_ui():
    """Start the Streamlit web interface."""
    app_path = str(PACKAGE_DIR / "ocr_app.py")
    try:
        cmd = [sys.executable, "-m", "streamlit", "run", app_path]
        subprocess.run(cmd, env=os.environ.copy(), check=True)
    except subprocess.CalledProcessError as e:
        # msg = f"Failed to start Streamlit: {e}"
        raise typer.Exit(1) from e
    except KeyboardInterrupt:
        logger.info("Shutting down...")


@cli.command()
def api(
    host: str = typer.Option("0.0.0.0", help="Host to bind to"),
    port: int = typer.Option(8000, help="Port to bind to"),
    log_level: str = typer.Option("info", help="Log level"),
    reload: bool = typer.Option(False, help="Enable auto-reload on file changes"),
):
    """Start the Docler API server."""
    try:
        # Lazy import to avoid unnecessary dependencies if not using the API
        import uvicorn

        from docler_api.main import app

        uvicorn.run(
            app,
            host=host,
            port=port,
            log_level=log_level,
            reload=reload,
        )
    except ImportError as e:
        msg = f"Failed to import API components: {e}. Is the 'server' extra installed?"
        logger.exception(msg)
        raise typer.Exit(1) from e
    except KeyboardInterrupt:
        logger.info("Shutting down...")


@cli.command()
def mcp(
    transport: TransportType = typer.Option(  # noqa: B008
        TransportType.STDIO, help="Transport protocol to use"
    ),
    host: str = typer.Option("0.0.0.0", help="Host to bind the server to"),
    port: int = typer.Option(8000, help="Port to bind the server to"),
):
    """Start the MCP server."""
    try:
        # Lazy import to avoid unnecessary dependencies if not using MCP
        from docler_mcp.server import mcp as mcp_server

        if transport == TransportType.STDIO:
            mcp_server.run(transport=transport.value)
        else:
            mcp_server.run(transport=transport.value, host=host, port=port)
    except ImportError as e:
        msg = f"Failed to import MCP components: {e}. Is the 'mcp' extra installed?"
        logger.exception(msg)
        raise typer.Exit(1) from e
    except KeyboardInterrupt:
        logger.info("Shutting down...")


def main():
    """Main entry point."""
    cli()


if __name__ == "__main__":
    main()



"""Module providing BGE embedding generation."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, ClassVar

from docler.configs.embedding_configs import BGEEmbeddingConfig
from docler.embeddings.base import EmbeddingProvider


if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    import numpy as np


class BGEEmbeddings(EmbeddingProvider[BGEEmbeddingConfig]):
    """Local embeddings using BGE models."""

    Config = BGEEmbeddingConfig

    NAME = "BGE"
    REQUIRED_PACKAGES: ClassVar = {"FlagEmbedding"}
    dimensions: ClassVar[int] = 1536  # OpenAI-compatible

    def __init__(self, model: str = "BAAI/bge-large-en-v1.5"):
        from FlagEmbedding import FlagModel

        self.model = FlagModel(model, use_fp16=True)  # pyright: ignore

    async def embed_stream(
        self,
        texts: AsyncIterator[str],
        batch_size: int = 8,
    ) -> AsyncIterator[np.ndarray]:
        batch: list[str] = []

        async for text in texts:
            batch.append(text)
            if len(batch) >= batch_size:
                embeddings = await asyncio.to_thread(self.model.encode, batch)
                for embedding in embeddings:
                    yield embedding
                batch = []

        if batch:
            embeddings = await asyncio.to_thread(self.model.encode, batch)
            for embedding in embeddings:
                yield embedding



"""Vector store implementation for document and text chunk storage."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Literal


if TYPE_CHECKING:
    from mkdown import TextChunk


Metric = Literal["cosine", "euclidean", "dot"]


class BaseVectorDB(ABC):
    """Abstract interface for vector databases that handle both storage and retrieval."""

    def __init__(self, vector_store_id: str):
        self.vector_store_id = vector_store_id

    @abstractmethod
    async def add_chunks(
        self,
        chunks: list[TextChunk],
    ) -> list[str]:
        """Add text chunks with metadata.

        Args:
            chunks: List of text chunks to add

        Returns:
            List of IDs for the stored chunks
        """

    @abstractmethod
    async def query(
        self,
        query: str,
        k: int = 4,
        filters: dict[str, Any] | None = None,
    ) -> list[tuple[TextChunk, float]]:
        """Find similar texts for a query.

        Args:
            query: Query text to search for
            k: Number of results to return
            filters: Optional filters to apply to results

        Returns:
            List of (text, score, metadata) tuples
        """

    @abstractmethod
    async def delete_chunk(self, chunk_id: str) -> bool:
        """Delete a chunk by ID.

        Args:
            chunk_id: ID of chunk to delete

        Returns:
            True if chunk was deleted, False otherwise
        """



import base64
from typing import Any

import anyenv
from mkdown import Image, create_page_break


def process_response(result: list[Any], api_key: str) -> tuple[list[str], list[Image]]:
    from llama_index.core.constants import DEFAULT_BASE_URL

    content_parts: list[str] = []
    images: list[Image] = []
    pages = result[0]["pages"]
    job_id = result[0]["job_id"]
    for page_num, page in enumerate(pages, start=1):
        if page.get("md"):
            comment = create_page_break(next_page=page_num)
            content_parts.append(comment)
            content_parts.append(page["md"])
        for img in page.get("images", []):
            image_count = len(images)
            id_ = f"img-{image_count}"
            asset_name = img["name"]
            asset_url = (
                f"{DEFAULT_BASE_URL}/api/parsing/job/{job_id}/result/image/{asset_name}"
            )
            headers = {"Authorization": f"Bearer {api_key}"}
            response = anyenv.get_bytes_sync(asset_url, headers=headers)
            img_data = base64.b64encode(response).decode("utf-8")
            img_type = "png"
            if "." in asset_name:
                extension = asset_name.split(".")[-1].lower()
                if extension in ["jpg", "jpeg", "png", "gif", "webp", "svg"]:
                    img_type = extension

            filename = f"{id_}.{img_type}"
            mime = f"image/{img_type}"
            image = Image(id=id_, content=img_data, mime_type=mime, filename=filename)
            images.append(image)
    return content_parts, images



"""FastAPI server for docler document conversion library."""

from __future__ import annotations

from typing import Any

from fastapi import FastAPI, File, Form, UploadFile
from fastapi.middleware.cors import CORSMiddleware
from mkdown import Document

from docler import __version__ as docler_version
from docler.configs.chunker_configs import ChunkerConfig
from docler.configs.converter_configs import ConverterConfig
from docler.models import ChunkedDocument, PageMetadata
from docler_api import routes


app = FastAPI(
    title="Docler API",
    description="API for document conversion using docler",
    version=docler_version,
)

# Add CORS middleware for cross-origin requests
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/")
async def root() -> dict[str, str]:
    """API root endpoint."""
    return {"message": "Welcome to Docler API", "version": docler_version}


@app.post("/api/convert")
async def api_convert_document(
    file: UploadFile = File(..., description="The document file to convert"),  # noqa: B008
    config: str = Form(..., description="Converter configuration"),
    pdf_password: str | None = Form(None, description="Password for encrypted PDF files"),
) -> Document:
    """Convert a document file to markdown."""
    return await routes.convert_document(file, config, pdf_password)


@app.get("/api/converters")
async def api_list_converters() -> dict[str, list[dict[str, Any]]]:
    """List all available converters."""
    return await routes.list_converters()


@app.get("/api/chunkers")
async def api_list_chunkers() -> dict[str, list[dict[str, Any]]]:
    """List all available chunkers."""
    return await routes.list_chunkers()


@app.post("/api/chunk")
async def api_chunk_document(
    file,
    converter_config: ConverterConfig,
    chunker_config: ChunkerConfig,
    pdf_password: str | None = None,
) -> ChunkedDocument:
    """Convert and chunk a document."""
    return await routes.chunk_document(
        file, converter_config, chunker_config, pdf_password
    )


@app.post("/api/pdf/metadata")
async def api_get_pdf_metadata(
    file: UploadFile = File(..., description="The PDF file to analyze"),  # noqa: B008
    pdf_password: str | None = Form(None, description="Password for encrypted PDF files"),
) -> PageMetadata:
    """Get PDF metadata including page count and document information."""
    return await routes.get_pdf_metadata(file, pdf_password)


# Additional endpoints for monitoring
@app.get("/health")
async def health_check() -> dict[str, str]:
    """Health check endpoint."""
    return {"status": "ok"}


@app.get("/ready")
async def readiness_check() -> dict[str, str]:
    """Readiness check endpoint."""
    return {"status": "ready"}


if __name__ == "__main__":
    import uvicorn

    uvicorn.run(app, host="0.0.0.0", port=8000)



"""Utility functions for the Streamlit app."""

from __future__ import annotations

from typing import TYPE_CHECKING

import streamlit as st


if TYPE_CHECKING:
    from mkdown import Document, TextChunk

    from docler.common_types import SupportedLanguage


LANGUAGES: list[SupportedLanguage] = ["en", "de", "fr", "es", "zh"]


def display_chunk_preview(chunk: TextChunk, expanded: bool = False):
    """Display a tabbed preview of a text chunk.

    Args:
        chunk: TextChunk to display
        expanded: Whether the expander should be initially expanded
    """
    header_text = f"Chunk {chunk.chunk_index + 1}"
    if chunk.metadata.get("header"):
        header_text += f" - {chunk.metadata['header']}"
    header_text += f" ({len(chunk.content)} chars)"
    with st.expander(header_text, expanded=expanded):
        tabs = ["Raw", "Rendered", "Debug Info", "Images"]
        raw_tab, rendered_tab, debug_tab, images_tab = st.tabs(tabs)
        with raw_tab:
            st.code(chunk.content, language="markdown")
        with rendered_tab:
            st.markdown(chunk.content)
        with debug_tab:
            debug_info = {
                "Chunk Index": chunk.chunk_index,
                "Source": chunk.source_doc_id,
                "Images": len(chunk.images),
                **chunk.metadata,
            }
            st.json(debug_info)

        with images_tab:
            if not chunk.images:
                st.info("No images in this chunk")
            else:
                for image in chunk.images:
                    data_url = image.to_base64_url()
                    st.markdown(f"**ID:** {image.id}")
                    if image.filename:
                        st.markdown(f"**Filename:** {image.filename}")
                    st.markdown(f"**MIME Type:** {image.mime_type}")
                    st.image(data_url)
                    st.divider()


def display_document_preview(doc: Document):
    """Display a tabbed preview of a document.

    Args:
        doc: Document to display
    """
    tabs = ["Raw Markdown", "Rendered", "Images"]
    raw_tab, rendered_tab, images_tab = st.tabs(tabs)
    with raw_tab:
        st.markdown(f"```markdown\n{doc.content}\n```")
    with rendered_tab:
        st.markdown(doc.content)
    with images_tab:
        if not doc.images:
            st.info("No images extracted")
        else:
            for image in doc.images:
                data_url = image.to_base64_url()
                st.markdown(f"**ID:** {image.id}")
                if image.filename:
                    st.markdown(f"**Filename:** {image.filename}")
                st.markdown(f"**MIME Type:** {image.mime_type}")
                st.image(data_url)
                st.divider()



"""Main Streamlit application for document processing."""

from __future__ import annotations

from typing import TYPE_CHECKING

import streamlit as st

from docler_streamlit.state import SessionState
from docler_streamlit.step1_conversion import show_step_1
from docler_streamlit.step2_correct import show_step_2
from docler_streamlit.step3_chunking import show_step_3
from docler_streamlit.step4_vectorstore import show_step_4


if TYPE_CHECKING:
    from mkdown import Document


def main():
    """Main Streamlit app."""
    st.title("Document Processing Pipeline")
    state = SessionState.get()
    with st.sidebar:
        st.title("Navigation")
        st.button("Reset App", on_click=state.reset)
        st.write(f"Current step: {state.step}")
        if state.uploaded_file_name:
            st.write(f"File: {state.uploaded_file_name}")
        if state.document:
            doc: Document = state.document
            st.write("Document Info:")
            st.write(f"- Title: {doc.title or 'Untitled'}")
            st.write(f"- Images: {len(doc.images)}")
            st.write(f"- Length: {len(doc.content)} chars")
        if state.chunks:
            st.write(f"- Chunks: {len(state.chunks)}")
            if state.vector_store_id:
                st.write(f"- Vector Store: {state.vector_store_id}")

    if state.step == 1:
        show_step_1()
    elif state.step == 2:  # noqa: PLR2004
        show_step_2()
    elif state.step == 3:  # noqa: PLR2004
        show_step_3()
    elif state.step == 4:  # noqa: PLR2004
        show_step_4()


if __name__ == "__main__":
    from streambricks import run

    run(main)



"""Module providing streaming embedding generation."""

from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING, ClassVar

from docler.configs.embedding_configs import OpenAIEmbeddingConfig
from docler.embeddings.base import EmbeddingProvider
from docler.utils import get_api_key


if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    import numpy as np


class OpenAIEmbeddings(EmbeddingProvider[OpenAIEmbeddingConfig]):
    """OpenAI embeddings with fallback to httpx if openai package not available."""

    Config = OpenAIEmbeddingConfig

    NAME = "OpenAI"
    REQUIRED_PACKAGES: ClassVar = {"openai", "httpx"}
    dimensions: ClassVar[int] = 1536

    def __init__(self, api_key: str | None = None, model: str = "text-embedding-3-small"):
        self.api_key = api_key or get_api_key("OPENAI_API_KEY")
        self.model = model
        self.use_openai = find_spec("openai")

    async def embed_stream(
        self,
        texts: AsyncIterator[str],
        batch_size: int = 8,
    ) -> AsyncIterator[np.ndarray]:
        """Embeddings iterator."""
        import numpy as np

        batch: list[str] = []

        async for text in texts:
            batch.append(text)
            if len(batch) >= batch_size:
                embeddings = (
                    await self._get_embeddings_official(batch)
                    if self.use_openai
                    else await self._get_embeddings_rest(batch)
                )
                for embedding in embeddings:
                    yield np.array(embedding, dtype=np.float32)
                batch = []

        if batch:
            embeddings = (
                await self._get_embeddings_official(batch)
                if self.use_openai
                else await self._get_embeddings_rest(batch)
            )
            for embedding in embeddings:
                yield np.array(embedding, dtype=np.float32)

    async def _get_embeddings_official(
        self,
        texts: list[str],
    ) -> list[list[float]]:
        """Get embeddings using official OpenAI client."""
        import openai

        client = openai.AsyncClient(api_key=self.api_key)
        response = await client.embeddings.create(model=self.model, input=texts)
        return [item.embedding for item in response.data]

    async def _get_embeddings_rest(
        self,
        texts: list[str],
    ) -> list[list[float]]:
        """Get embeddings using httpx."""
        import anyenv

        url = "https://api.openai.com/v1/embeddings"
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        data = {"input": texts, "model": self.model}
        result = await anyenv.post_json(
            url,
            json_data=data,
            headers=headers,
            return_type=dict,
        )
        return [item["embedding"] for item in result["data"]]



"""Configuration models for document and image annotators."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import Field
from schemez import ModelIdentifier  # noqa: TC002

from docler.common_types import DEFAULT_ANNOTATOR_MODEL, DEFAULT_IMAGE_ANNOTATOR_MODEL
from docler.provider import ProviderConfig


if TYPE_CHECKING:
    from docler.annotators.base import Annotator


DEFAULT_DOC_SYSTEM_PROMPT = """
You are an expert document analyzer that extracts meaningful metadata.
For each document or text chunk, extract:
1. Main topics (3-5 categories)
2. Key entities (people, organizations, locations, products)
3. Keywords (5-10 important terms)

Format your response as structured data that can be parsed as JSON.
"""

DEFAULT_DOC_PROMPT_TEMPLATE = """
Complete context:
    {context}

Please analyze and describe this text chunk:
    {chunk}
"""

DEFAULT_IMAGE_SYSTEM_PROMPT = """
Analyze images in detail. For each image, provide:
1. A detailed description of what's visible
2. Key objects/people present
3. Any text content visible in the image
4. Image type (photo, chart, diagram, illustration, etc.)
5. Dominant colors and visual elements

Format your response as structured data that can be parsed as JSON.
"""

DEFAULT_IMAGE_PROMPT_TEMPLATE = """
Analyze this image with ID {image_id}{filename_info}.
Describe what you see and extract key information.
"""


class BaseAnnotatorConfig(ProviderConfig):
    """Base configuration for annotators."""


class AIDocumentAnnotatorConfig(BaseAnnotatorConfig):
    """Configuration for AI-based document and chunk annotator."""

    type: Literal["ai_document"] = Field(default="ai_document", init=False)
    """Type discriminator for AI document annotator."""

    model: ModelIdentifier = DEFAULT_ANNOTATOR_MODEL
    """LLM model to use for annotation."""

    system_prompt: str = DEFAULT_DOC_SYSTEM_PROMPT
    """System prompt for the annotation task."""

    user_prompt: str = DEFAULT_DOC_PROMPT_TEMPLATE
    """Template for the annotation prompt."""

    max_context_length: int = Field(default=1500, ge=1)
    """Maximum length of context for annotation."""

    batch_size: int = Field(default=5, ge=1)
    """Number of chunks to process in parallel."""

    def get_provider(self) -> Annotator:
        """Get the annotator instance."""
        from docler.annotators.ai_document_annotator import AIDocumentAnnotator

        return AIDocumentAnnotator(**self.get_config_fields())


class AIImageAnnotatorConfig(BaseAnnotatorConfig):
    """Configuration for AI-based image annotator."""

    type: Literal["ai_image"] = Field(default="ai_image", init=False)
    """Type discriminator for AI image annotator."""

    model: ModelIdentifier = DEFAULT_IMAGE_ANNOTATOR_MODEL
    """Vision model to use for image annotation."""

    system_prompt: str = DEFAULT_IMAGE_SYSTEM_PROMPT
    """System prompt for the image analysis task."""

    user_prompt: str = DEFAULT_IMAGE_PROMPT_TEMPLATE
    """User prompt template for image analysis."""

    batch_size: int = Field(default=3, ge=1)
    """Number of images to process concurrently."""

    def get_provider(self) -> Annotator:
        """Get the annotator instance."""
        from docler.annotators.ai_image_annotator import AIImageAnnotator

        return AIImageAnnotator(**self.get_config_fields())


# Union type for annotator configs
AnnotatorConfig = Annotated[
    AIDocumentAnnotatorConfig | AIImageAnnotatorConfig,
    Field(discriminator="type"),
]



"""Streamlit application package for Docler document processing."""

from docler_streamlit.app import main
from docler_streamlit.chunk_app import main as chunk_app
from docler_streamlit.ocr_app import main as ocr_app

__all__ = ["chunk_app", "main", "ocr_app"]



"""Document converter using DataLab's API."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Literal

from mkdown import Document
from upathtools import read_path, to_upath

from docler.configs.converter_configs import DataLabConfig
from docler.converters.base import DocumentConverter
from docler.converters.datalab_provider.utils import get_response, process_response
from docler.log import get_logger
from docler.pdf_utils import shift_page_range
from docler.utils import get_api_key


if TYPE_CHECKING:
    from schemez import MimeType

    from docler.common_types import PageRangeString, StrPath, SupportedLanguage


logger = get_logger(__name__)

Mode = Literal["marker", "table_rec", "ocr", "layout"]

# See https://www.datalab.to/app/reference
# https://www.datalab.to/openapi.json


class DataLabConverter(DocumentConverter[DataLabConfig]):
    """Document converter using DataLab's API."""

    Config = DataLabConfig

    NAME = "datalab"
    SUPPORTED_MIME_TYPES: ClassVar[set[str]] = {
        # PDF
        "application/pdf",
        # Images
        "image/png",
        "image/jpeg",
        "image/webp",
        "image/gif",
        "image/tiff",
        "image/jpg",
        # Office Documents
        "application/msword",  # doc
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document",  # docx
        # Presentations
        "application/vnd.ms-powerpoint",  # ppt
        "application/vnd.openxmlformats-officedocument.presentationml.presentation",  # pptx  # noqa: E501
    }

    def __init__(
        self,
        languages: list[SupportedLanguage] | None = None,
        *,
        page_range: PageRangeString | None = None,
        api_key: str | None = None,
        force_ocr: bool = False,
        use_llm: bool = False,
    ):
        """Initialize the DataLab converter.

        Args:
            page_range: Page range(s) to extract, like "1-5,7-10" (1-based)
            api_key: DataLab API key.
            languages: Languages to use for OCR.
            force_ocr: Whether to force OCR on every page.
            use_llm: Whether to use LLM for enhanced accuracy.
        """
        super().__init__(languages=languages, page_range=page_range)
        self.api_key = api_key or get_api_key("DATALAB_API_KEY")
        self.force_ocr = force_ocr
        self.use_llm = use_llm
        self.add_page_breaks = True

    @property
    def price_per_page(self) -> float:
        """Price per page in USD."""
        return 0.003 if self.use_llm else 0.0015

    async def _convert_path_async(
        self,
        file_path: StrPath,
        mime_type: MimeType,
    ) -> Document:
        """Convert a file using DataLab's API.

        Args:
            file_path: Path to the file to process.
            mime_type: MIME type of the file.

        Returns:
            Converted document.

        Raises:
            ValueError: If conversion fails.
        """
        path = to_upath(file_path)
        form = {"output_format": "markdown", "paginate": self.add_page_breaks}
        data = await read_path(path, mode="rb")
        files = {"file": (path.name, data, mime_type)}
        if self.languages:
            form["langs"] = ",".join(self.languages)
        if self.force_ocr:
            form["force_ocr"] = "true"
        if self.use_llm:
            form["use_llm"] = "true"
        if self.page_range:
            # DataLab expects 0-based indexes in page_range
            rng = shift_page_range(self.page_range, shift=-1) if self.page_range else None
            form["page_range"] = rng
        result = await get_response(form, files, self.api_key)
        md_content, images = process_response(result)
        return Document(
            content=md_content.strip(),
            images=images,
            title=path.stem,
            source_path=str(path),
            mime_type=mime_type,
        )


if __name__ == "__main__":
    import anyenv

    pdf_path = "src/docler/resources/pdf_sample.pdf"
    converter = DataLabConverter()
    result = anyenv.run_sync(converter.convert_file(pdf_path))
    print(result)



"""Converter configuration."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal

# from docling.datamodel.pipeline_options import (
#     EasyOcrOptions,
#     OcrMacOptions,
#     RapidOcrOptions,
#     TesseractCliOcrOptions,
#     TesseractOcrOptions,
# )
from pydantic import Field, HttpUrl, SecretStr  # noqa: TC002
from pydantic_settings import SettingsConfigDict
from schemez import ModelIdentifier  # noqa: TC002

from docler.common_types import (
    DEFAULT_CONVERTER_MODEL,
    PageRangeString,  # noqa: TC001
    SupportedLanguage,  # noqa: TC001
)
from docler.provider import ProviderConfig


if TYPE_CHECKING:
    from docler.converters.base import DocumentConverter


ConverterShorthand = Literal[
    "docling",
    "docling_remote",
    "marker",
    "mistral",
    "llamaparse",
    "datalab",
    "azure",
    "llm",
]

DoclingEngine = Literal[
    "easy_ocr", "tesseract_cli_ocr", "tesseract_ocr", "ocr_mac", "rapid_ocr"
]

UpstageOCRType = Literal["auto", "force"]
UpstageOutputFormat = Literal["markdown", "text", "html"]
UpstageCategory = Literal["figure", "chart", "table", "paragraph"]

AzureModel = Literal[
    "prebuilt-read",
    "prebuilt-layout",
    "prebuilt-idDocument",
    "prebuilt-receipt",
]

AzureFeatureFlag = Literal[
    "ocrHighResolution",
    "languages",
    "barcodes",
    "formulas",
    "keyValuePairs",
    "styleFont",
    "queryFields",
]


LlamaParseMode = Literal[
    "parse_page_without_llm",
    "parse_page_with_llm",
    "parse_page_with_lvm",
    "parse_page_with_agent",
    "parse_document_with_llm",
]


MarkerLLMProvider = Literal["gemini", "ollama", "vertex", "claude"]


LLM_SYSTEM_PROMPT = """
You are a OCR engine that can parse PDF documents & images into markdown formatted text.
"""


LLM_USER_PROMPT = """
Convert this PDF document to markdown format
Preserve the original formatting and structure where possible.
Include any important tables or lists.
Describe any images you see in brackets.
Indicate page breaks with a specially formatted JSON comment,
like <!-- docler:page_break {"next_page":2} --> for page 2.
Only respond with the content, nothing else.
"""


def default_languages() -> set[SupportedLanguage]:
    return {"en"}


class BaseConverterConfig(ProviderConfig):
    """Base configuration for document converters."""

    page_range: PageRangeString | None = None
    """Page range(s) to extract, like "1-5,7-10" (1-based)."""

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        raise NotImplementedError


class DoclingConverterConfig(BaseConverterConfig):
    """Configuration for docling-based converter."""

    type: Literal["docling"] = Field("docling", init=False)
    """Type discriminator for docling converter."""

    languages: set[SupportedLanguage] = Field(default_factory=default_languages)
    """List of supported languages for the converter."""

    image_scale: float = Field(default=2.0, gt=0)
    """Scale factor for image resizing."""

    generate_images: bool = True
    """Whether to generate images."""

    ocr_engine: (
        DoclingEngine
        # | EasyOcrOptions
        # | TesseractCliOcrOptions
        # | TesseractOcrOptions
        # | OcrMacOptions
        # | RapidOcrOptions
    ) = "easy_ocr"
    """OCR engine to use."""

    model_config = SettingsConfigDict(env_prefix="DOCLING_")

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.docling_provider import DoclingConverter

        return DoclingConverter(**self.get_config_fields())


class MarkItDownConfig(BaseConverterConfig):
    """Configuration for MarkItDown-based converter."""

    type: Literal["markitdown"] = Field("markitdown", init=False)
    """Type discriminator for MarkItDown converter."""

    languages: set[SupportedLanguage] = Field(default_factory=default_languages)
    """List of supported languages for the converter."""

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.markitdown_provider import MarkItDownConverter

        return MarkItDownConverter(**self.get_config_fields())


class DataLabConfig(BaseConverterConfig):
    """Configuration for DataLab-based converter."""

    type: Literal["datalab"] = Field("datalab", init=False)
    """Type discriminator for DataLab converter."""

    languages: set[SupportedLanguage] = Field(default_factory=default_languages)
    """List of supported languages for the converter."""

    api_key: SecretStr | None = None
    """DataLab API key. If None, will try env var DATALAB_API_KEY."""

    force_ocr: bool = False
    """Whether to force OCR on every page."""

    use_llm: bool = False
    """Whether to use LLM for enhanced accuracy."""

    max_pages: int | None = Field(default=None, ge=1)
    """Maximum number of pages to process."""

    model_config = SettingsConfigDict(env_prefix="DATALAB_")

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.datalab_provider import DataLabConverter

        return DataLabConverter(**self.get_config_fields())


class LLMConverterConfig(BaseConverterConfig):
    """Configuration for LLM-based converter."""

    type: Literal["llm"] = Field("llm", init=False)
    """Type discriminator for LLM converter."""

    languages: set[SupportedLanguage] = Field(default_factory=default_languages)
    """List of supported languages for the converter."""

    model: ModelIdentifier = DEFAULT_CONVERTER_MODEL
    """LLM model to use."""

    system_prompt: str = LLM_SYSTEM_PROMPT
    """Optional system prompt to guide conversion."""

    user_prompt: str = LLM_USER_PROMPT
    """Custom prompt for the conversion task."""

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.llm_provider import LLMConverter

        return LLMConverter(**self.get_config_fields())


class MistralConfig(BaseConverterConfig):
    """Configuration for Mistral-based converter."""

    type: Literal["mistral"] = Field("mistral", init=False)
    """Type discriminator for Mistral converter."""

    languages: set[SupportedLanguage] = Field(default_factory=default_languages)
    """List of supported languages for the converter."""

    api_key: SecretStr | None = None
    """Mistral API key. If None, will try env var MISTRAL_API_KEY."""

    model_config = SettingsConfigDict(env_prefix="MISTRAL_")

    # right now there only is one model
    # ocr_model: str = "mistral-ocr-latest"
    # """Mistral OCR model to use."""

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.mistral_provider import MistralConverter

        return MistralConverter(**self.get_config_fields())


class LlamaParseConfig(BaseConverterConfig):
    """Configuration for LlamaParse-based converter."""

    type: Literal["llamaparse"] = Field("llamaparse", init=False)
    """Type discriminator for LlamaParse converter."""

    languages: set[SupportedLanguage] = Field(default_factory=default_languages)
    """List of supported languages for the converter."""

    api_key: SecretStr | None = None
    """LlamaParse API key. Falls back to LLAMAPARSE_API_KEY env var."""

    adaptive_long_table: bool = False
    """Whether to use adaptive long table."""

    parse_mode: LlamaParseMode = "parse_page_with_llm"
    """Parse mode, defaults to "parse_page_with_llm"."""

    skip_diagonal_text: bool = False
    """Whether to skip diagonal text."""

    disable_ocr: bool = False
    """Whether to disable OCR for images."""

    continuous_mode: bool = True
    """Whether to use continuous mode."""

    html_tables: bool = False
    """Whether to output HTML tables instead of markdown."""

    model_config = SettingsConfigDict(env_prefix="LLAMAPARSE_")

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.llamaparse_provider import LlamaParseConverter

        return LlamaParseConverter(**self.get_config_fields())


class AzureConfig(BaseConverterConfig):
    """Configuration for Azure Document Intelligence converter."""

    type: Literal["azure"] = Field("azure", init=False)
    """Type discriminator for Azure converter."""

    languages: set[SupportedLanguage] = Field(default_factory=default_languages)
    """List of supported languages for the converter."""

    endpoint: str | None = None
    """Azure endpoint URL. Falls back to AZURE_DOC_INTELLIGENCE_ENDPOINT envvar."""

    api_key: SecretStr | None = None
    """Azure API key. Falls back to AZURE_DOC_INTELLIGENCE_KEY env var."""

    model: AzureModel = "prebuilt-layout"
    """Pre-trained model to use."""

    additional_features: set[AzureFeatureFlag] = Field(default_factory=set)
    """Optional add-on capabilities like BARCODES, FORMULAS, OCR_HIGH_RESOLUTION etc."""

    model_config = SettingsConfigDict(env_prefix="LLAMAPARSE_")

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.azure_provider import AzureConverter

        return AzureConverter(**self.get_config_fields())


class MarkerConfig(BaseConverterConfig):
    """Configuration for Marker-based converter."""

    type: Literal["marker"] = Field("marker", init=False)
    """Type discriminator for Marker converter."""

    dpi: int = Field(default=192, gt=0)
    """DPI setting for image extraction."""

    llm_provider: MarkerLLMProvider | None = None
    """Language model provider to use for OCR."""

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.marker_provider import MarkerConverter

        return MarkerConverter(**self.get_config_fields())


class UpstageConfig(BaseConverterConfig):
    """Configuration for Upstage Document AI converter."""

    type: Literal["upstage"] = Field("upstage", init=False)
    """Type discriminator for Upstage converter."""

    api_key: SecretStr | None = None
    """Upstage API key. Falls back to UPSTAGE_API_KEY env var."""

    base_url: HttpUrl = HttpUrl("https://api.upstage.ai/v1/document-ai/document-parse")
    """API endpoint URL."""

    # model: str = "document-parse"
    # """Model name for document parsing."""

    ocr: UpstageOCRType = "auto"
    """OCR mode ('auto' or 'force')."""

    base64_categories: set[UpstageCategory] = Field(
        default_factory=lambda: {"figure", "chart"}  # type: ignore
    )
    """Element categories to encode in base64."""

    model_config = SettingsConfigDict(env_prefix="UPSTAGE_")

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.upstage_provider import UpstageConverter

        return UpstageConverter(**self.get_config_fields())


class DoclingRemoteConfig(BaseConverterConfig):
    """Configuration for remote Docling service."""

    type: Literal["docling_remote"] = Field("docling_remote", init=False)
    """Type discriminator for remote docling converter."""

    languages: set[SupportedLanguage] = Field(default_factory=default_languages)
    """List of supported languages for the converter."""

    api_key: SecretStr | None = None
    """Optional API key for auth."""

    endpoint: HttpUrl = HttpUrl("http://localhost:5001")
    """Base URL of the Docling service."""

    ocr_engine: DoclingEngine = "easy_ocr"
    """OCR engine to use."""

    pdf_backend: str = "dlparse_v4"
    """PDF backend to use."""

    table_mode: Literal["fast", "accurate"] = "fast"
    """Table mode to use."""

    force_ocr: bool = False
    """Whether to force OCR even on digital documents."""

    image_scale: float = Field(default=2.0, gt=0)
    """Scale factor for image resolution."""

    do_table_structure: bool = True
    """Extract table structure."""

    do_code_enrichment: bool = False
    """Enable code extraction."""

    do_formula_enrichment: bool = False
    """Enable formula extraction."""

    do_picture_classification: bool = False
    """Enable picture classification."""

    do_picture_description: bool = False
    """Enable picture description."""

    model_config = SettingsConfigDict(env_prefix="DOCLING_REMOTE_")

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.docling_remote_provider import DoclingRemoteConverter

        return DoclingRemoteConverter(**self.get_config_fields())


class AggregatedConverterConfig(BaseConverterConfig):
    """Configuration for the aggregated converter."""

    type: Literal["aggregated"] = Field(default="aggregated", init=False)
    """Type discriminator for aggregated converter."""

    converters: list[ConverterConfig] = Field(default_factory=list)
    """List of converter configurations to use."""

    mime_preferences: dict[str, str] = Field(default_factory=dict)
    """Mapping of MIME types or extensions to preferred converter names."""

    def get_provider(self) -> DocumentConverter:
        """Get the converter instance."""
        from docler.converters.aggregated_converter import AggregatedConverter

        return AggregatedConverter.from_config(self)


# class UnstructuredConverterConfig(BaseConverterConfig):
#     """Configuration for Unstructured-based converter."""

#     typ e: Literal["unstructured"] = "unstructured"
#     """Type discriminator for Unstructured converter."""

#     languages: set[SupportedLanguage] = Field(default_factory=lambda: {"en"})
#     """List of supported languages for the converter."""

#     api_key: str | None = None
#     """API key for Unstructured (optional for local processing)."""

#     strategy: str = "hi_res"
#     """Strategy for document processing (hi_res, fast, etc.)."""

#     extract_images: bool = True
#     """Whether to extract and include images."""

#     extract_tables: bool = True
#     """Whether to extract tables as images."""

#     local_mode: bool = True
#     """Whether to use local processing or API."""

#     def get_provider(self) -> DocumentConverter:
#         """Get the converter instance."""
#         from docler.converters.unstructured_provider import UnstructuredConverter

#         return UnstructuredConverter(**self.get_config_fields())


ConverterConfig = Annotated[
    DataLabConfig
    | DoclingConverterConfig
    | DoclingRemoteConfig
    | LLMConverterConfig
    | MarkItDownConfig
    | MistralConfig
    | LlamaParseConfig
    | AzureConfig
    | UpstageConfig
    # | UnstructuredConverterConfig
    | AggregatedConverterConfig
    | MarkerConfig,
    Field(discriminator="type"),
]



"""AI-based markdown chunking implementation."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from mkdown import TextChunk


if TYPE_CHECKING:
    from mkdown import Document

    from docler.chunkers.ai_chunker.models import Chunk


def add_line_numbers(text: str) -> str:
    """Add line numbers to text."""
    lines = text.splitlines()
    return "\n".join(f"{i + 1:3d} | {line}" for i, line in enumerate(lines))


def create_text_chunk(
    doc: Document,
    chunk: Chunk,
    chunk_idx: int,
    extra_metadata: dict[str, Any] | None = None,
) -> TextChunk:
    """Create a TextChunk from chunk definition."""
    lines = doc.content.splitlines()
    chunk_lines = lines[chunk.start_row - 1 : chunk.end_row]
    chunk_text = "\n".join(chunk_lines)
    base = extra_metadata or {}
    metadata = {**base, "keywords": chunk.keywords, "references": chunk.references}
    chunk_images = [i for i in doc.images if i.filename and i.filename in chunk_text]
    return TextChunk(
        content=chunk_text,
        source_doc_id=doc.source_path or "",
        chunk_index=chunk_idx,
        images=chunk_images,
        metadata=metadata,
    )



"""Utils for docling-remote provider."""

from __future__ import annotations

import base64
import re
from typing import Any

from mkdown import Image, create_image_reference


def process_response(document: dict[str, Any]) -> tuple[str, list[Image]]:
    content = document["md_content"]
    images: list[Image] = []

    if "![" in content:
        img_pattern = r"!\[([^]]*)\]\(data:image/([^;]+);base64,([^)]+)\)"
        image_counter = 0

        def replace_image(match) -> str:
            nonlocal image_counter
            alt_text = match.group(1)
            img_type = match.group(2)
            img_data = match.group(3)

            image_id = f"img-{image_counter}"
            filename = f"{image_id}.{img_type}"
            image_counter += 1
            content = base64.b64decode(img_data)
            mime = f"image/{img_type}"
            image = Image(id=image_id, content=content, mime_type=mime, filename=filename)
            images.append(image)
            return create_image_reference(alt_text or image_id, filename)

        content = re.sub(img_pattern, replace_image, content)
    return content, images



"""Utilities for generating diffs between texts."""

from __future__ import annotations

from difflib import unified_diff
import importlib.util


def generate_unified_diff(original: str, corrected: str) -> str:
    """Generate a unified diff between original and corrected text.

    Args:
        original: Original text content
        corrected: Corrected text content

    Returns:
        String containing the unified diff
    """
    diff_lines = unified_diff(
        original.splitlines(),
        corrected.splitlines(),
        fromfile="original",
        tofile="corrected",
        lineterm="",
    )

    return "\n".join(diff_lines)


def generate_semantic_diff(original: str, corrected: str) -> str:
    """Generate a semantic diff between original and corrected text.

    Args:
        original: Original text content
        corrected: Corrected text content

    Returns:
        JSON string containing the semantic diff
    """
    import anyenv
    from diff_match_patch import diff_match_patch

    dmp = diff_match_patch()
    diffs = dmp.diff_main(original, corrected)
    dmp.diff_cleanupSemantic(diffs)

    # Convert to a JSON-friendly format
    result = []
    for op, text in diffs:
        op_name = {-1: "delete", 0: "equal", 1: "insert"}[op]
        result.append({"operation": op_name, "text": text})

    return anyenv.dump_json(result)


def generate_html_diff(original: str, corrected: str) -> str:
    """Generate an HTML visualization of the differences.

    Args:
        original: Original text content
        corrected: Corrected text content

    Returns:
        HTML string with highlighted differences
    """
    from diff_match_patch import diff_match_patch

    dmp = diff_match_patch()
    diffs = dmp.diff_main(original, corrected)
    dmp.diff_cleanupSemantic(diffs)

    return dmp.diff_prettyHtml(diffs)


def generate_all_diffs(original: str, corrected: str) -> dict[str, str]:
    """Generate all available diff formats between texts.

    Args:
        original: Original text content
        corrected: Corrected text content

    Returns:
        Dictionary with all available diff formats
    """
    result = {"unified_diff": generate_unified_diff(original, corrected)}

    if importlib.util.find_spec("diff_match_patch") is not None:
        result["semantic_diff"] = generate_semantic_diff(original, corrected)
        result["html_diff"] = generate_html_diff(original, corrected)

    return result



"""Chunkers configuration for the Streamlit app."""

from __future__ import annotations

from docler.chunkers.ai_chunker import AIChunker
from docler.chunkers.llamaindex_chunker import LlamaIndexChunker
from docler.chunkers.markdown_chunker import MarkdownChunker


CHUNKERS = {"Markdown": MarkdownChunker, "LlamaIndex": LlamaIndexChunker, "AI": AIChunker}



"""OpenAI Vector Store implementation."""

from __future__ import annotations

from typing import Any


def convert_filters(filters: dict[str, Any]) -> Any:  # type: ignore
    """Convert standard filters to OpenAI's filter format.

    Args:
        filters: Dictionary of filters

    Returns:
        OpenAI-compatible filter object
    """
    filter_conditions = []
    for key, value in filters.items():
        if isinstance(value, list):
            # Handle list of values (OR condition)
            or_conditions = [
                {"key": key, "type": "eq", "value": item}
                for item in value
                if isinstance(item, str | int | float | bool)
            ]

            if or_conditions:
                filter_conditions.append({"type": "or", "filters": or_conditions})
        elif isinstance(value, str | int | float | bool):
            # Handle single value (equality)
            filter_conditions.append({"key": key, "type": "eq", "value": value})  # type: ignore
    if len(filter_conditions) > 1:
        return {"type": "and", "filters": filter_conditions}
    if filter_conditions:
        return filter_conditions[0]
    return {}



"""Step 2: Document correction/proof reading interface."""

from __future__ import annotations

import anyenv
import streambricks as sb
import streamlit as st

from docler.configs.processor_configs import LLMProofReaderConfig
from docler.log import get_logger
from docler_streamlit.state import SessionState


logger = get_logger(__name__)


def show_step_2():
    """Show document proof reading/correction screen (step 2)."""
    state = SessionState.get()
    st.header("Step 2: Document Correction")
    col1, col2, col3 = st.columns([1, 4, 1])
    with col1:
        st.button("← Back", on_click=state.prev_step)
    with col3:
        # When proceeding without corrections, copy original document
        def proceed_to_chunking():
            if state.document and not state.corrected_document:
                state.corrected_document = state.document
            state.next_step()

        st.button("Proceed to Chunking →", on_click=proceed_to_chunking)

    if not state.document:
        st.warning("No document to process. Please go back and convert a document first.")
        return

    st.subheader("Proof Reader Configuration")
    if not state.processor_config:
        state.processor_config = LLMProofReaderConfig()
    state.processor_config = sb.model_edit(state.processor_config)
    if st.button("Process Document"):
        with st.spinner("Processing document with AI proof reader..."):
            try:
                processor = state.processor_config.get_provider()
                processed_doc = anyenv.run_sync(processor.process(state.document))
                state.corrected_document = processed_doc
                st.success("Document successfully processed!")
            except Exception as e:
                st.error(f"Processing failed: {e}")
                logger.exception("Document processing failed")

    if state.corrected_document:
        st.subheader("Document Corrections")
        original_content = state.document.content
        corrected_content = state.corrected_document.content
        if original_content == corrected_content:
            st.info("No corrections were made to the document.")
        else:
            from st_diff_viewer import diff_viewer

            diff_viewer(
                original_content,
                corrected_content,
                split_view=True,
                title="OCR Corrections",
            )
            if "proof_reading" in state.corrected_document.metadata:
                proof_metadata = state.corrected_document.metadata["proof_reading"]
                col1, col2 = st.columns(2)
                with col1:
                    count = proof_metadata.get("corrections_count", 0)
                    st.metric("Total Corrections", count)
                with col2:
                    if count:
                        lines = original_content.splitlines()
                        percentage = min(100, (count / len(lines)) * 100)
                        st.metric("% Lines Corrected", f"{percentage:.1f}%")
                if proof_metadata.get("corrections"):
                    st.subheader("Line Corrections")
                    for correction in proof_metadata["corrections"]:
                        with st.expander(f"Line {correction.get('line_number')}"):
                            st.code(correction.get("corrected"), language="markdown")

        st.subheader("Processed Document Preview")
        tabs = st.tabs(["Rendered Content", "Raw Markdown"])
        with tabs[0]:
            st.markdown(state.corrected_document.content)
        with tabs[1]:
            st.code(state.corrected_document.content, language="markdown")
    else:
        st.info(
            "You can proceed to chunking without corrections, "
            "or use the processor to correct common OCR errors."
        )
        with st.expander("View Original Document"):
            st.markdown(state.document.content)


if __name__ == "__main__":
    from streambricks import run

    # For testing purposes
    run(show_step_2)



"""Vector store configuration."""

from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import Field, HttpUrl, SecretStr  # noqa: TC002
from pydantic.functional_validators import model_validator

from docler.provider import ProviderConfig


if TYPE_CHECKING:
    from docler.vector_db.base_manager import VectorManagerBase


Metric = Literal["cosine", "euclidean", "dotproduct"]
PineconeRegion = Literal[
    "us-east-1", "us-west-2", "eu-west-1", "europe-west4", "us-central1", "eastus2"
]
PineconeCloud = Literal["aws", "gcp", "azure"]


VectorDBShorthand = Literal["chroma", "qdrant", "pinecone"]


class BaseVectorStoreConfig(ProviderConfig):
    """Base configuration for vector stores."""

    def get_provider(self) -> VectorManagerBase:
        """Get the converter instance."""
        raise NotImplementedError


class ChromaConfig(BaseVectorStoreConfig):
    """Configuration for ChromaDB vector store."""

    type: Literal["chroma"] = Field(default="chroma", init=False)

    persist_directory: str | None = None
    """Where to persist the database."""


class QdrantConfig(BaseVectorStoreConfig):
    """Configuration for Qdrant vector store."""

    type: Literal["qdrant"] = Field(default="qdrant", init=False)

    location: str | None = None
    """Path to local Qdrant storage. If None, uses memory."""

    url: HttpUrl | None = None
    """URL for Qdrant server. If set, location is ignored."""

    api_key: SecretStr | None = None
    """API key for Qdrant cloud."""

    prefer_grpc: bool = True
    """Whether to prefer gRPC over HTTP."""

    @model_validator(mode="after")
    def validate_connection(self) -> QdrantConfig:
        """Ensure either location or url is set, but not both."""
        if self.location and self.url:
            msg = "Cannot specify both location and url"
            raise ValueError(msg)
        if self.api_key and not self.url:
            msg = "API key only valid with url"
            raise ValueError(msg)
        return self


class KdbAiConfig(BaseVectorStoreConfig):
    """Configuration for KDB.AI vector store."""

    type: Literal["kdbai"] = Field(default="kdbai", init=False)

    endpoint: str | None = None
    """Server endpoint to connect to."""

    api_key: SecretStr | None = None
    """API Key for authentication."""

    mode: Literal["rest", "qipc"] | None = None
    """Implementation method used for the session."""

    table_name: str = "vectors"
    """Name of the table to store vectors."""

    index_type: Literal["flat", "hnsw"] = "hnsw"
    """Type of index to use."""

    @model_validator(mode="after")
    def validate_config(self) -> KdbAiConfig:
        """Validate configuration."""
        if not self.endpoint and not self.api_key:
            msg = "Must specify either endpoint or api_key"
            raise ValueError(msg)
        return self


class PineconeConfig(BaseVectorStoreConfig):
    """Configuration for Pinecone vector store."""

    type: Literal["pinecone"] = Field(default="pinecone", init=False)
    """Type identifier for Pinecone."""

    api_key: SecretStr | None = None
    """Pinecone API key."""

    environment: str = "us-west1-gcp"
    """Pinecone environment."""

    cloud: PineconeCloud = "aws"
    """Cloud provider (aws, gcp, azure)."""

    region: PineconeRegion = "us-east-1"
    """Cloud region."""

    dimension: int = Field(default=1536, gt=0)
    """Vector dimension."""

    metric: Metric = "cosine"
    """Distance metric for similarity search."""


VectorStoreConfig = Annotated[
    ChromaConfig | QdrantConfig | KdbAiConfig | PineconeConfig,
    Field(discriminator="type"),
]



"""Document converter using pypdf."""

from __future__ import annotations

import io
from typing import TYPE_CHECKING

# import fitz  # PyMuPDF
from pypdf import PdfReader, PdfWriter

from docler.log import get_logger
from docler.models import PageDimensions, PageMetadata


if TYPE_CHECKING:
    from docler.common_types import PageRangeString


logger = get_logger(__name__)


def parse_page_range(page_range: PageRangeString, shift: int = 0) -> set[int]:
    """Convert a page range string to a set of page numbers.

    Args:
        page_range: String like "1-5,7,9-11" or None.
        shift: Amount to shift page numbers by (e.g., -1 to convert 1-based to 0-based)

    Returns:
        Set of page numbers (shifted by specified amount)

    Raises:
        ValueError: If the page range format is invalid.
    """
    if shift:
        page_range = shift_page_range(page_range, shift)

    pages: set[int] = set()
    try:
        for part in page_range.split(","):
            if "-" in part:
                start, end = map(int, part.split("-"))
                pages.update(range(start, end + 1))
            else:
                pages.add(int(part))
    except ValueError as e:
        msg = f"Invalid page range format: {page_range}. Expected format: '1-5,7,9-11'"
        raise ValueError(msg) from e
    else:
        return pages


def shift_page_range(page_range: PageRangeString, shift: int = 0) -> PageRangeString:
    """Shift page numbers in a page range string by the specified amount.

    Args:
        page_range: Page range string like "1-5,7,9-11"
        shift: Amount to shift page numbers by (e.g., -1 to convert 1-based to 0-based)

    Returns:
        Shifted page range string

    Raises:
        ValueError: If the page range format is invalid
    """
    parts = []
    try:
        for part in page_range.split(","):
            if "-" in part:
                start, end = map(int, part.split("-"))
                if start + shift < 0 or end + shift < 0:
                    msg = f"Invalid shift {shift} for page range {page_range}"
                    raise ValueError(msg)  # noqa: TRY301
                parts.append(f"{start + shift}-{end + shift}")
            else:
                page = int(part)
                if page + shift < 0:
                    msg = f"Invalid shift {shift} for page {page}"
                    raise ValueError(msg)  # noqa: TRY301
                parts.append(str(page + shift))
    except ValueError as e:
        if "Invalid shift" in str(e):
            raise
        msg = f"Invalid page range format: {page_range}. Expected format: '1-5,7,9-11'"
        raise ValueError(msg) from e

    return ",".join(parts)


def decrypt_pdf(data: bytes, password: str | None) -> bytes:
    """Decrypt password-protected PDF and return decrypted bytes.

    Args:
        data: Encrypted PDF file content as bytes
        password: Password to decrypt the PDF, or None to try empty password

    Returns:
        Decrypted PDF content as bytes

    Raises:
        ValueError: If decryption fails or password is incorrect
    """
    with io.BytesIO(data) as pdf_io, io.BytesIO() as output:
        try:
            reader = PdfReader(pdf_io)
            if not reader.is_encrypted:
                return data  # Already decrypted

            # Try empty password first if no password provided
            if password is None:
                if reader.decrypt(""):
                    # Successfully decrypted with empty password
                    writer = PdfWriter()
                    for page in reader.pages:
                        writer.add_page(page)
                    writer.write(output)
                    return output.getvalue()
                msg = "PDF is encrypted and requires a password"
                raise ValueError(msg)  # noqa: TRY301

            # Try provided password
            if not reader.decrypt(password):
                msg = "Incorrect password for encrypted PDF"
                raise ValueError(msg)  # noqa: TRY301

            writer = PdfWriter()
            for page in reader.pages:
                writer.add_page(page)
            writer.write(output)
            return output.getvalue()
        except Exception as e:
            if "Incorrect password" in str(e) or "requires a password" in str(e):
                raise
            msg = f"Failed to decrypt PDF: {e}"
            raise ValueError(msg) from e


def extract_pdf_pages(
    data: bytes, page_range: PageRangeString | None, password: str | None = None
) -> bytes:
    """Extract specific pages from a PDF file and return as new PDF.

    Args:
        data: Source PDF file content as bytes
        page_range: String like "1-5,7,9-11" or None for all pages. 1-based.
        password: Password for encrypted PDFs

    Returns:
        New PDF containing only specified pages as bytes

    Raises:
        ValueError: If page range is invalid, PDF data cannot be processed,
                    or decryption fails
    """
    with io.BytesIO(data) as pdf_io, io.BytesIO() as output:
        try:
            reader = PdfReader(pdf_io)

            # Handle encrypted PDFs
            if reader.is_encrypted:
                if password is not None:
                    # Try provided password
                    if not reader.decrypt(password):
                        msg = "Incorrect password for encrypted PDF"
                        raise ValueError(msg)  # noqa: TRY301
                # Try empty password first
                elif not reader.decrypt(""):
                    msg = "PDF is encrypted but no password provided"
                    raise ValueError(msg)  # noqa: TRY301

            pages = (
                parse_page_range(page_range, shift=-1)
                if page_range
                else range(len(reader.pages))
            )
            writer = PdfWriter()
            for i in pages:
                if 0 <= i < len(reader.pages):
                    writer.add_page(reader.pages[i])
            writer.write(output)
            return output.getvalue()
        except Exception as e:
            if "encrypted" in str(e).lower():
                raise
            msg = f"Failed to extract pages from PDF: {e}"
            raise ValueError(msg) from e


def get_pdf_info(data: bytes, password: str | None = None) -> PageMetadata:
    """Get PDF metadata including page count, dimensions, and file info.

    Args:
        data: PDF file content as bytes
        password: Password for encrypted PDFs

    Returns:
        PageMetadata model containing PDF information

    Raises:
        ValueError: If PDF data cannot be processed or decryption fails
    """
    with io.BytesIO(data) as pdf_io:
        try:
            reader = PdfReader(pdf_io)

            # Handle encrypted PDFs
            is_encrypted = False
            if reader.is_encrypted:
                if password is not None:
                    # Try provided password
                    if not reader.decrypt(password):
                        msg = "Incorrect password for encrypted PDF"
                        raise ValueError(msg)  # noqa: TRY301
                # Try empty password first (many PDFs are encrypted with empty password)
                elif reader.decrypt(""):
                    # Successfully decrypted with empty password, treat as not encrypted
                    is_encrypted = False
                else:
                    # Return basic info for truly encrypted PDF without decryption
                    return PageMetadata(
                        page_count=0,
                        file_size=len(data),
                        is_encrypted=True,
                        page_dimensions=[],
                        title="",
                        author="",
                    )

            # Basic info
            page_count = len(reader.pages)
            file_size = len(data)

            # Page dimensions (in points)
            page_dimensions = []
            for page in reader.pages:
                media_box = page.mediabox
                width = float(media_box.width)
                height = float(media_box.height)
                page_dimensions.append(PageDimensions(width=width, height=height))

            # Document metadata
            metadata = reader.metadata
            title = metadata.title or "" if metadata else ""
            author = metadata.author or "" if metadata else ""

            return PageMetadata(
                page_count=page_count,
                file_size=file_size,
                is_encrypted=is_encrypted,
                page_dimensions=page_dimensions,
                title=title,
                author=author,
            )
        except Exception as e:
            if "Incorrect password" in str(e):
                raise
            msg = f"Failed to get PDF info: {e}"
            raise ValueError(msg) from e



"""Utilities."""

from __future__ import annotations

import base64
import io
from io import BytesIO
import re
from typing import TYPE_CHECKING, Literal, overload


if TYPE_CHECKING:
    import PIL.Image
    from PIL.ImageFile import ImageFile

    from docler.common_types import StrPath


def pil_to_bytes(image: PIL.Image.Image) -> bytes:
    """Convert PIL image to bytes in its native format."""
    img_byte_arr = BytesIO()
    image.save(img_byte_arr, format=image.format or "JPEG")
    return img_byte_arr.getvalue()


def get_mime_from_pil(image: PIL.Image.Image, fallback: str = "JPEG") -> str:
    """Get MIME type from PIL image format."""
    fmt = image.format or fallback
    return f"image/{fmt.lower()}"


@overload
def check_mime(
    path: StrPath,
    *,
    allowed_mime_types: set[str] | None = None,
    raise_if_none_found: Literal[True],
) -> str: ...


@overload
def check_mime(
    path: StrPath,
    *,
    allowed_mime_types: set[str] | None = None,
    raise_if_none_found: Literal[False] = False,
) -> str | None: ...


def check_mime(
    path: StrPath,
    *,
    allowed_mime_types: set[str] | None = None,
    raise_if_none_found: bool = False,
) -> str | None:
    """Return and optionally check MIME type for a path.

    Determines the MIME type of a file or extension and raises an error
    if the type cannot be determined or if it is not allowed.

    Args:
        path: Path to the file (like "test.jpg", or just the extension starting with ".")
        allowed_mime_types: Set of allowed MIME types.
        raise_if_none_found: Whether to raise an error if the type cant be determined.
    """
    import mimetypes

    path_str = str(path)
    if path_str.startswith("."):
        path_str = f"file{path_str}"
    mime = mimetypes.guess_type(path_str)[0]
    if mime is None and raise_if_none_found:
        msg = f"Could not determine MIME type for {path}"
        raise ValueError(msg)
    if allowed_mime_types and mime not in allowed_mime_types:
        msg = f"Invalid MIME type: {mime}. Allowed types: {allowed_mime_types}"
        raise ValueError(msg)
    return mime


def decode_base64_to_image(encoded_string, image_format="PNG") -> ImageFile:
    """Decode a base64 string to an image."""
    from PIL import Image

    try:
        image_data = base64.b64decode(encoded_string)
        return Image.open(io.BytesIO(image_data))
    except Exception as e:  # noqa: BLE001
        msg = f"Failed to decode image: {e!s}"
        raise ValueError(msg)  # noqa: B904


def encode_image_to_base64(image, image_format="WEBP", quality=20):
    """Encode an image to base64 string."""
    buffer = io.BytesIO()
    image.save(buffer, format=image_format, quality=quality)
    return base64.b64encode(buffer.getvalue()).decode("utf-8")


def png_to_webp(content: str) -> str:
    """Convert PNG images to WebP format in Markdown content."""
    pattern = re.compile(r"!\[Image\]\(data:image/png;base64,([^)]*)\)")
    matches = pattern.findall(content)

    for match in matches:
        try:
            png_image = decode_base64_to_image(match, "PNG")

            if png_image.format != "PNG":
                continue

            if png_image.width > 1080:  # noqa: PLR2004
                ratio = png_image.height / png_image.width
                webp_image = png_image.resize((1080, int(1080 * ratio)))
            else:
                webp_image = png_image

            webp_encoded_string = encode_image_to_base64(webp_image, "WEBP", quality=20)
            content = content.replace(
                f"data:image/png;base64,{match}",
                f"data:image/webp;base64,{webp_encoded_string}",
            )

        except Exception:  # noqa: BLE001
            continue

    return content


def get_api_key(env_var: str) -> str:
    """Get environment variable, throw if not set."""
    import os

    key = os.getenv(env_var)
    if not key:
        msg = f"Required environment variable {env_var} not set"
        raise ValueError(msg)
    return key


def add_line_numbers(text: str) -> str:
    """Add line numbers to text."""
    lines = text.splitlines()
    return "\n".join(f"{i + 1:5d} | {line}" for i, line in enumerate(lines))



"""Base classes for text chunking implementations."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, ClassVar

from docler.provider import BaseProvider


if TYPE_CHECKING:
    from docler.configs.annotator_configs import BaseAnnotatorConfig
    from docler.models import ChunkedDocument


class Annotator[TConfig](ABC, BaseProvider[TConfig]):
    """Base class for chunk annotation processors."""

    Config: ClassVar[type[BaseAnnotatorConfig]]

    @abstractmethod
    async def annotate(
        self,
        chunked_doc: ChunkedDocument,
    ) -> ChunkedDocument:
        """Annotate a chunked document with additional metadata.

        Args:
            chunked_doc: Chunked document containing the original content and chunks

        Returns:
            Chunked document with annotated chunks
        """
        raise NotImplementedError



from __future__ import annotations

from typing import Literal

from fastmcp import FastMCP
from mcp.types import ToolAnnotations
from upath import UPath

from docler.common_types import SupportedLanguage
from docler.converters.base import DocumentConverter


ConverterShorthand = Literal[
    "docling",
    "marker",
    "mistral",
    "llamaparse",
    "datalab",
    "azure",
    "llm",
    "markitdown",
    "upstage",
]


mcp = FastMCP("phil65")


@mcp.tool(
    annotations=ToolAnnotations(
        title="Convert to Markdown",
        readOnlyHint=False,
        idempotentHint=True,
    )
)
async def convert_to_markdown(
    # ctx: Context,
    source_path: str,
    target_folder: str,
    provider: ConverterShorthand = "mistral",
    languages: list[SupportedLanguage] | None = None,
    page_range: str | None = None,
) -> str:
    """Convert a PDF to markdown and images.

    Args:
        source_path: Path to the source document (PDF, powerpoint, images, ...)
        target_folder: Path where to save the resulting markdown and images
        provider: The provider to use for conversion.
        languages: Used languages in the document, if known.
        page_range: 1-based page range, for example "1,4-7,9-10"
    """
    try:
        target = UPath(target_folder).expanduser()
        source = UPath(source_path).expanduser()
        target.mkdir(parents=True, exist_ok=True)
        converters: list[type[DocumentConverter]] = (
            DocumentConverter.get_available_providers()
        )
        converter_cls = next(i for i in converters if provider == i.NAME)
        converter = converter_cls(languages=languages, page_range=page_range)
        # await ctx.info(f"Starting conversion for {source_path} using {provider!r}...")
        document = await converter.convert_file(source)
        # await ctx.info(f"Exporting to {target}...")
        await document.export_to_directory(target)
        files_str = "\n- ".join(str(i) for i in target.iterdir())
    except FileNotFoundError as e:
        return f"Error: {e}. Is the path correct? Current cwd is {UPath.cwd()}."
    else:
        return f"Conversion successful. Files created in {target_folder}:\n{files_str}"


if __name__ == "__main__":
    import logging

    logging.basicConfig()
    mcp.run()



"""Streamlit app for markdown chunking visualization."""

from __future__ import annotations

import anyenv
from mkdown import Document
import streamlit as st

from docler.chunkers.markdown_chunker import MarkdownChunker
from docler_streamlit.utils import display_chunk_preview


SAMPLE_MARKDOWN = """# Introduction

This is a sample markdown document to test chunking.

## First Section

Some content here with an example image:
![example](image.png)

### Subsection

More detailed content...

## Second Section

Another block of text that demonstrates how chunking works.
"""


def main():
    """Main Streamlit app."""
    st.title("Markdown Chunker")
    with st.sidebar:
        st.header("Chunker Settings")
        min_size = st.number_input(
            "Minimum chunk size",
            min_value=50,
            max_value=1000,
            value=200,
            step=50,
            help="Minimum characters per chunk",
        )
        max_size = st.number_input(
            "Maximum chunk size",
            min_value=100,
            max_value=2000,
            value=1500,
            step=100,
            help="Maximum characters per chunk",
        )
        overlap = st.number_input(
            "Chunk overlap",
            min_value=0,
            max_value=200,
            value=50,
            step=10,
            help="Character overlap between chunks",
        )

    col1, col2 = st.columns([2, 1])
    with col1:
        st.subheader("Input Markdown")
    with col2:
        if st.button("Load Sample"):
            st.session_state.markdown_input = SAMPLE_MARKDOWN
    input_text = st.session_state.get("markdown_input", "")
    text = "Enter your markdown here:"
    markdown_input = st.text_area(text, value=input_text, height=200)

    if markdown_input:
        doc = Document(content=markdown_input, source_path="input.md")
        chunker = MarkdownChunker(
            min_chunk_size=min_size,
            max_chunk_size=max_size,
            chunk_overlap=overlap,
        )
        if st.button("Chunk Markdown"):
            with st.spinner("Processing..."):
                chunked = anyenv.run_sync(chunker.chunk(doc))
                st.subheader("Chunks")
                for chunk in chunked.chunks:
                    display_chunk_preview(chunk, expanded=True)


if __name__ == "__main__":
    from streambricks import run

    run(main)



"""Docler: main package.

Abstractions & Tools for OCR / document processing.
"""

from __future__ import annotations

from importlib.metadata import version

__version__ = version("docler")
__title__ = "Docler"

__author__ = "Philipp Temminghoff"
__author_email__ = "philipptemminghoff@googlemail.com"
__copyright__ = "Copyright (c) 2025 Philipp Temminghoff"
__license__ = "MIT"
__url__ = "https://github.com/phil65/docler"

from docler.converters.base import DocumentConverter
from docler.converters.dir_converter import Conversion, DirectoryConverter
from docler.converters.registry import ConverterRegistry

# Import providers
from docler.converters.aggregated_converter import AggregatedConverter
from docler.converters.azure_provider import AzureConverter
from docler.converters.datalab_provider import DataLabConverter
from docler.converters.docling_provider import DoclingConverter
from docler.converters.llamaparse_provider import LlamaParseConverter
from docler.converters.llm_provider import LLMConverter
from docler.converters.marker_provider import MarkerConverter
from docler.converters.markitdown_provider import MarkItDownConverter
from docler.converters.mistral_provider import MistralConverter
from docler.converters.upstage_provider import UpstageConverter

__all__ = [
    "AggregatedConverter",
    "AzureConverter",
    "Conversion",
    "ConverterRegistry",
    "DataLabConverter",
    "DirectoryConverter",
    "DoclingConverter",
    "DocumentConverter",
    "LLMConverter",
    "LlamaParseConverter",
    "MarkItDownConverter",
    "MarkerConverter",
    "MistralConverter",
    "UpstageConverter",
    "__version__",
]



"""Base class for vector database managers."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, ClassVar

from docler.provider import BaseProvider


if TYPE_CHECKING:
    from docler.configs.vector_db_configs import BaseVectorStoreConfig
    from docler.models import VectorStoreInfo
    from docler.vector_db.base import BaseVectorDB


class VectorManagerBase[TConfig](BaseProvider[TConfig], ABC):
    """Abstract base class for vector database managers."""

    Config: ClassVar[type[BaseVectorStoreConfig]]

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of this vector database provider."""

    @abstractmethod
    async def list_vector_stores(self) -> list[VectorStoreInfo]:
        """List all available vector stores for this provider."""

    @abstractmethod
    async def create_vector_store(self, name: str, **kwargs) -> BaseVectorDB:
        """Create a new vector store."""

    @abstractmethod
    async def get_vector_store(self, name: str, **kwargs) -> BaseVectorDB:
        """Get a connection to an existing vector store."""

    @abstractmethod
    async def delete_vector_store(self, name: str) -> bool:
        """Delete a vector store."""

    @abstractmethod
    async def close(self):
        """Close all vector store connections."""

    async def delete_all_vector_stores(self) -> bool:
        """Delete all vector stores."""
        stores = await self.list_vector_stores()
        for store in stores:
            await self.delete_vector_store(store.db_id)
        return True

    async def has_vector_store(self, name: str) -> bool:
        """Check if a vector store exists."""
        indexes = await self.list_vector_stores()
        index_names = [idx.name for idx in indexes]
        return name in index_names
